import ants

import numpy as np

def randomly_transform_image_data(reference_image,
                                  input_image_list,
                                  segmentation_image_list=None,
                                  number_of_simulations=10,
                                  transform_type='affine',
                                  sd_affine=0.02,
                                  deformation_transform_type="bspline",
                                  number_of_random_points=1000,
                                  sd_noise=10.0,
                                  number_of_fitting_levels=4,
                                  mesh_size=1,
                                  sd_smoothing=4.0,
                                  input_image_interpolator='linear',
                                  segmentation_image_interpolator='nearestNeighbor'):

    """
    Randomly transform image data (optional: with corresponding segmentations).

    Apply rigid, affine and/or deformable maps to an input set of training
    images.  The reference image domain defines the space in which this
    happens.

    Arguments
    ---------
    reference_image : ANTsImage
        Defines the spatial domain for all output images.  If the input images do
        not match the spatial domain of the reference image, we internally
        resample the target to the reference image.  This could have unexpected
        consequences.  Resampling to the reference domain is performed by testing
        using ants.image_physical_space_consistency then calling
        ants.resample_image_to_target with failure.

    input_image_list : list of lists of ANTsImages
        List of lists of input images to warp.  The internal list sets contain one
        or more images (per subject) which are assumed to be mutually aligned.  The
        outer list contains multiple subject lists which are randomly sampled to
        produce output image list.

    segmentation_image_list : list of ANTsImages
        List of segmentation images corresponding to the input image list (optional).

    number_of_simulations : integer
        Number of output images.

    transform_type : string
        One of the following options: "translation", "rigid", "scaleShear", "affine",
        "deformation", "affineAndDeformation".

    sd_affine : float
        Parameter dictating deviation amount from identity for random linear
        transformations.

    deformation_transform_type : string
        "bspline" or "exponential".

    number_of_random_points : integer
        Number of displacement points for the deformation field.

    sd_noise : float
        Standard deviation of the displacement field.

    number_of_fitting_levels : integer
        Number of fitting levels (bspline deformation only).

    mesh_size : int or n-D tuple
        Determines fitting resolution (bspline deformation only).

    sd_smoothing : float
        Standard deviation of the Gaussian smoothing in mm (exponential field only).

    input_image_interpolator : string
        One of the following options "linear", "gaussian", "bspline".

    segmentation_image_interpolator : string
        One of the following options "nearestNeighbor" or "genericLabel".

    Returns
    -------
    list of lists of transformed images

    Example
    -------
    >>> import ants
    >>> image1_list = list()
    >>> image1_list.append(ants.image_read(ants.get_ants_data("r16")))
    >>> image2_list = list()
    >>> image2_list.append(ants.image_read(ants.get_ants_data("r64")))
    >>> input_segmentations = list()
    >>> input_segmentations.append(ants.threshold_image(image1, "Otsu", 3))
    >>> input_segmentations.append(ants.threshold_image(image2, "Otsu", 3))
    >>> input_images = list()
    >>> input_images.append(image1_list)
    >>> input_images.append(image2_list)
    >>> data = antspynet.randomly_transform_image_data(image1,
    >>>     input_images, input_segmentations, sd_affine=0.02,
    >>>     transform_type = "affineAndDeformation" )
    """

    from ..utilities import histogram_warp_image_intensities

    def polar_decomposition(X):
         U, d, V = np.linalg.svd(X, full_matrices=False)
         P = np.matmul(U, np.matmul(np.diag(d), np.transpose(U)))
         Z = np.matmul(U, V)
         if np.linalg.det(Z) < 0:
             n = X.shape[0]
             reflection_matrix = np.identity(n)
             reflection_matrix[0,0] = -1.0
             Z = np.matmul(Z, reflection_matrix)
         return({"P" : P, "Z" : Z, "Xtilde" : np.matmul(P, Z)})

    def create_random_linear_transform(image,
                                       fixed_parameters,
                                       transform_type='affine',
                                       sd_affine=0.02):
        transform = ants.create_ants_transform(transform_type=
           "AffineTransform", precision='float', dimension=image.dimension)
        ants.set_ants_transform_fixed_parameters(transform, fixed_parameters)
        identity_parameters = ants.get_ants_transform_parameters(transform)
        random_epsilon = np.random.normal(loc=0, scale=sd_affine, size=len(identity_parameters))

        if transform_type == 'translation':
            random_epsilon[:(len(identity_parameters) - image.dimension)] = 0

        random_parameters = identity_parameters + random_epsilon
        random_matrix = np.reshape(
            random_parameters[:(len(identity_parameters) - image.dimension)],
            newshape=(image.dimension, image.dimension))
        decomposition = polar_decomposition(random_matrix)

        if transform_type == "rigid":
            random_matrix = decomposition['Z']
        elif transform_type == "affine":
            random_matrix = decomposition['Xtilde']
        elif transform_type == "scaleShear":
            random_matrix = decomposition['P']

        random_parameters[:(len(identity_parameters) - image.dimension)] = \
            np.reshape(random_matrix, newshape=(len(identity_parameters) - image.dimension))
        ants.set_ants_transform_parameters(transform, random_parameters)
        return(transform)

    def create_random_displacement_field_transform(image,
                                                   field_type="bspline",
                                                   number_of_random_points=1000,
                                                   sd_noise=10.0,
                                                   number_of_fitting_levels=4,
                                                   mesh_size=1,
                                                   sd_smoothing=4.0):
        displacement_field = ants.simulate_displacement_field(image,
            field_type=field_type, number_of_random_points=number_of_random_points,
            sd_noise=sd_noise, enforce_stationary_boundary=True,
            number_of_fitting_levels=number_of_fitting_levels, mesh_size=mesh_size,
            sd_smoothing=sd_smoothing)
        return(ants.transform_from_displacement_field(displacement_field))

    admissible_transforms = ("translation", "rigid", "scaleShear", "affine",
        "affineAndDeformation", "deformation")
    if not transform_type in admissible_transforms:
        raise ValueError("The specified transform is not a possible option.  Please see help menu.")

    # Get the fixed parameters from the reference image.

    fixed_parameters = ants.get_center_of_mass(reference_image)
    number_of_subjects = len(input_image_list)

    random_indices = np.random.choice(number_of_subjects, size=number_of_simulations, replace=True)

    simulated_image_list = list()
    simulated_segmentation_image_list = list()
    simulated_transforms = list()

    for i in range(number_of_simulations):
        single_subject_image_list = input_image_list[random_indices[i]]
        single_subject_segmentation_image = None
        if segmentation_image_list is not None:
            single_subject_segmentation_image = segmentation_image_list[random_indices[i]]

        if ants.image_physical_space_consistency(reference_image, single_subject_image_list[0]) is False:
            for j in range(len(single_subject_image_list)):
                single_subject_image_list.append(
                    ants.resample_image_to_target(single_subject_image_list[j], reference_image,
                        interp_type=input_image_interpolator))
            if single_subject_segmentation_image is not None:
                single_subject_segmentation_image = \
                    ants.resample_image_to_target(single_subject_segmentation_image, reference_image,
                        interp_type=segmentation_image_interpolator)

        transforms = list()

        if transform_type == 'deformation':
            deformable_transform = create_random_displacement_field_transform(
                reference_image, deformation_transform_type, number_of_random_points,
                sd_noise, number_of_fitting_levels, mesh_size, sd_smoothing)
            transforms.append(deformable_transform)
        elif transform_type == 'affineAndDeformation':
            deformable_transform = create_random_displacement_field_transform(
                reference_image, deformation_transform_type, number_of_random_points,
                sd_noise, number_of_fitting_levels, mesh_size, sd_smoothing)
            linear_transform = create_random_linear_transform(reference_image,
                fixed_parameters, 'affine', sd_affine)
            transforms.append(deformable_transform)
            transforms.append(linear_transform)
        else:
            linear_transform = create_random_linear_transform(reference_image,
                fixed_parameters, transform_type, sd_affine)
            transforms.append(linear_transform)

        simulated_transforms.append(ants.compose_ants_transforms(transforms))

        single_subject_simulated_image_list = list()
        for j in range(len(single_subject_image_list)):
            single_subject_image = single_subject_image_list[j]
            single_subject_simulated_image_list.append(ants.apply_ants_transform_to_image(
                simulated_transforms[i], single_subject_image,
                reference=reference_image))

        simulated_image_list.append(single_subject_simulated_image_list)

        if single_subject_segmentation_image is not None:
            simulated_segmentation_image_list.append(ants.apply_ants_transform_to_image(
                simulated_transforms[i], single_subject_segmentation_image,
                reference=reference_image, interpolation="nearestneighbor"))

    if segmentation_image_list is None:
        return({'simulated_images' : simulated_image_list,
                'simulated_transforms' : simulated_transforms})
    else:
        return({'simulated_images' : simulated_image_list,
                'simulated_segmentation_images' : simulated_segmentation_image_list,
                'simulated_transforms' : simulated_transforms})



import numpy as np
import tensorflow as tf
import ants

def tid_neural_image_assessment(image,
                                mask=None,
                                patch_size=101,
                                stride_length=None,
                                padding_size=0,
                                dimensions_to_predict=0,
                                antsxnet_cache_directory=None,
                                which_model="tidsQualityAssessment",
                                verbose=False):

    """
    Perform MOS-based assessment of an image.

    Use a ResNet architecture to estimate image quality in 2D or 3D using subjective
    QC image databases described in

    https://www.sciencedirect.com/science/article/pii/S0923596514001490

    or

    https://doi.org/10.1109/TIP.2020.2967829

    where the image assessment is either "global", i.e., a single number or an image
    based on the specified patch size.  In the 3-D case, neighboring slices are used
    for each estimate.  Note that parameters should be kept as consistent as possible
    in order to enable comparison.  Patch size should be roughly 1/12th to 1/4th of
    image size to enable locality. A global estimate can be gained by setting
    patch_size = "global".

    Arguments
    ---------
    image : ANTsImage (2-D or 3-D)
        input image.

    mask : ANTsImage (2-D or 3-D)
        optional mask for designating calculation ROI.

    patch_size : integer
        prime number of patch_size.  101 is good.  Otherwise, choose "global" for a single
        global estimate of quality.

    stride_length : integer or vector of image dimension length
        optional value to speed up computation (typically less than patch size).

    padding_size : positive or negative integer or vector of image dimension length
        de(padding) to remove edge effects.

    dimensions_to_predict : integer or vector
        if image dimension is 3, this parameter specifies which dimensions should be used for
        prediction.  If more than one dimension is specified, the results are averaged.

    antsxnet_cache_directory : string
        Destination directory for storing the downloaded template and model weights.
        Since these can be resused, if is None, these data will be downloaded to
        ~/.keras/ANTsXNet/.

    which_model : string
        model type e.g. string tidsQualityAssessment, koniqMS, koniqMS2 or koniqMS3 where
        the former predicts mean opinion score (MOS) and MOS standard deviation and
        the latter koniq models predict mean opinion score (MOS) and sharpness.

    verbose : boolean
        Print progress to the screen.

    Returns
    -------
    List of QC results predicting both both human rater's mean and standard
    deviation of the MOS ("mean opinion scores") or sharpness depending on the
    selected network.  Both aggregate and spatial scores are returned, the latter
    in the form of an image.

    Example
    -------
    >>> image = ants.image_read(ants.get_data("r16"))
    >>> mask = ants.get_mask(image)
    >>> tid = tid_neural_image_assessment(image, mask=mask, patch_size=101, stride_length=7)
    """

    from ..utilities import get_pretrained_network
    from ..utilities import pad_or_crop_image_to_size
    from ..utilities import extract_image_patches
    from ..utilities import reconstruct_image_from_patches

    def is_prime(n):
        if n == 2 or n == 3:
            return True
        if n < 2 or n % 2 == 0:
            return False
        if n < 9:
            return True
        if n%3 == 0:
            return False
        r = int(n ** 0.5)
        f = 5
        while f <= r:
            if n % f == 0:
                return False
            if n % (f + 2) == 0:
                return False
            f += 6
        return True

    valid_models = ("tidsQualityAssessment", "koniqMS", "koniqMS2", "koniqMS3")
    if not which_model in valid_models:
        raise ValueError("Please pass valid model")

    if antsxnet_cache_directory == None:
        antsxnet_cache_directory = "ANTsXNet"

    if verbose == True:
        print("Neural QA:  retreiving model and weights.")

    is_koniq = "koniq" in which_model
    model_and_weights_file_name = get_pretrained_network(which_model, antsxnet_cache_directory=antsxnet_cache_directory)
    tid_model = tf.keras.models.load_model(model_and_weights_file_name, compile=False)

    padding_size_vector = padding_size
    if isinstance(padding_size, int):
        padding_size_vector = np.repeat(padding_size, image.dimension)
    elif len(padding_size) == 1:
        padding_size_vector = np.repeat(padding_size[0], image.dimension)

    if isinstance(dimensions_to_predict, int):
        dimensions_to_predict = (dimensions_to_predict,)

    padded_image_size = image.shape + padding_size_vector
    padded_image = pad_or_crop_image_to_size(image, padded_image_size)

    number_of_channels = 3

    if stride_length is None and patch_size != "global":
        stride_length = round(min(patch_size) / 2)
        if image.dimension == 3:
            stride_length = (stride_length, stride_length, 1)

    ###############
    #
    #  Global
    #
    ###############

    if patch_size == 'global':
        if which_model == "tidsQualityAssessment":
            evaluation_image = ants.iMath(padded_image, "Normalize") * 255

        if is_koniq:
            evaluation_image = ants.iMath(padded_image, "Normalize") * 2.0 - 1.0

        if image.dimension == 2:
            batchX = np.zeros((1, evaluation_image.shape, number_of_channels))
            for k in range(3):
                batchX[0,:,:,k] = evaluation_image.numpy()
            predicted_data = tid_model.predict(batchX, verbose=verbose)

            if which_model == "tidsQualityAssessment":
                return_dict = {'MOS' : None,
                               'MOS.standardDeviation' : None,
                               'MOS.mean' : predicted_data[0, 0],
                               'MOS.standardDeviationMean' : predicted_data[0, 1]
                              }
                return(return_dict)

            elif is_koniq:
                return_dict = {'MOS.mean' : predicted_data[0, 0],
                               'sharpness.mean' : predicted_data[0, 1]
                              }
                return(return_dict)

        elif image.dimension == 3:
            mos_mean = 0
            mos_standard_deviation = 0

            x = tuple(range(image.dimension))
            for d in range(len(dimensions_to_predict)):
                not_padded_image_size = list(padded_image_size)
                del(not_padded_image_size[dimensions_to_predict[d]])
                batchX = np.zeros((padded_image_size[dimensions_to_predict[d]],
                    (*tuple(not_padded_image_size)), number_of_channels))

                batchX[0,:,:,0] = (ants.slice_image(evaluation_image, axis=0, idx=dimensions_to_predict[d])).numpy()
                batchX[0,:,:,1] = (ants.slice_image(evaluation_image, axis=0, idx=dimensions_to_predict[d])).numpy()
                batchX[0,:,:,2] = (ants.slice_image(evaluation_image, axis=1, idx=dimensions_to_predict[d])).numpy()
                for i in range(1, padded_image_size[dimensions_to_predict[d]] - 1):
                    batchX[i,:,:,0] = (ants.slice_image(evaluation_image, axis=i-1, idx=dimensions_to_predict[d])).numpy()
                    batchX[i,:,:,1] = (ants.slice_image(evaluation_image, axis=i  , idx=dimensions_to_predict[d])).numpy()
                    batchX[i,:,:,2] = (ants.slice_image(evaluation_image, axis=i+1, idx=dimensions_to_predict[d])).numpy()
                batchX[padded_image_size[dimensions_to_predict[d]],:,:,0] = (
                    ants.slice_image(evaluation_image, axis=padded_image_size[dimensions_to_predict[d]]-1,
                        idx=dimensions_to_predict[d])).numpy()
                batchX[padded_image_size[dimensions_to_predict[d]],:,:,1] = (
                    ants.slice_image(evaluation_image, axis=padded_image_size[dimensions_to_predict[d]],
                        idx=dimensions_to_predict[d])).numpy()
                batchX[padded_image_size[dimensions_to_predict[d]],:,:,2] = (
                    ants.slice_image(evaluation_image, axis=padded_image_size[dimensions_to_predict[d]],
                        idx=dimensions_to_predict[d])).numpy()

                predicted_data = tid_model.predict(batchX, verbose=verbose)
                mos_mean += predicted_data[0, 0]
                mos_standard_deviation += predicted_data[0, 1]

            mos_mean /= len(dimensions_to_predict)
            mos_standard_deviation /= len(dimensions_to_predict)
            if which_model == "tidsQualityAssessment":
                return_dict = {'MOS.mean' : mos_mean,
                               'MOS.standardDeviationMean' : mos_standard_deviation
                              }
                return(return_dict)
            elif is_koniq:
                return_dict = {'MOS.mean' : mos_mean,
                               'sharpness.mean' : mos_standard_deviation
                              }
                return(return_dict)

    ###############
    #
    #  Patchwise
    #
    ###############

    else:

        evaluation_image = padded_image

        if not is_prime(patch_size):
            print("patch_size should be a prime number:  13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97...")

        stride_length_vector = stride_length
        if isinstance(stride_length, int):
            if image.dimension == 2:
                stride_length_vector = (stride_length, stride_length)
        elif len(stride_length) == 1:
            if image.dimension == 2:
                stride_length_vector = (stride_length[0], stride_length[0])

        patch_size_vector = (patch_size, patch_size)

        if image.dimension == 2:
            dimensions_to_predict = (1,)

        permutations = list()

        mos = image * 0
        mos_standard_deviation = image * 0

        for d in range(len(dimensions_to_predict)):
            if image.dimension == 3:
                permutations.append((0, 1, 2))
                permutations.append((0, 2, 1))
                permutations.append((1, 2, 0))

                if dimensions_to_predict[d] == 0:
                    patch_size_vector = (patch_size, patch_size, number_of_channels)
                    if isinstance(stride_length, int):
                        stride_length_vector = (stride_length, stride_length, 1)
                elif dimensions_to_predict[d] == 1:
                    patch_size_vector = (patch_size, number_of_channels, patch_size)
                    if isinstance(stride_length, int):
                        stride_length_vector = (stride_length, 1, stride_length)
                elif dimensions_to_predict[d] == 2:
                    patch_size_vector = (number_of_channels, patch_size, patch_size)
                    if isinstance(stride_length, int):
                        stride_length_vector = (1, stride_length, stride_length)
                else:
                    raise ValueError("dimensions_to_predict elements should be 1, 2, and/or 3 for 3-D image.")

            patches = extract_image_patches(evaluation_image, patch_size=patch_size_vector,
            stride_length=stride_length_vector, return_as_array=False)

            batchX = np.zeros((len(patches), patch_size, patch_size, number_of_channels))

            is_good_patch = np.repeat(False, len(patches))
            for i in range(len(patches)):
                if patches[i].var() > 0:
                    is_good_patch[i] = True
                    patch_image = patches[i]
                    patch_image = patch_image - patch_image.min()

                    if patch_image.max() > 0:
                        if which_model == "tidsQualityAssessment":
                            patch_image = patch_image / patch_image.max() * 255
                        elif is_koniq:
                            patch_image = patch_image / patch_image.max() * 2.0 - 1.0

                    if image.dimension == 2:
                        for j in range(number_of_channels):
                            batchX[i,:,:,j] = patch_image
                    elif image.dimension == 3:
                        batchX[i,:,:,:] = np.transpose(np.squeeze(patch_image.numpy()), permutations[dimensions_to_predict[d]])

            good_batchX = batchX[is_good_patch,:,:,:]
            predicted_data = tid_model.predict(good_batchX, verbose=verbose)

            patches_mos = list()
            patches_mos_standard_deviation = list()

            zero_patch_image = patch_image * 0

            count = 0
            for i in range(len(patches)):
                if is_good_patch[i]:
                    patches_mos.append(zero_patch_image + predicted_data[count, 0])
                    patches_mos_standard_deviation.append(zero_patch_image + predicted_data[count, 1])
                    count += 1
                else:
                    patches_mos.append(zero_patch_image)
                    patches_mos_standard_deviation.append(zero_patch_image)

            mos += pad_or_crop_image_to_size(reconstruct_image_from_patches(
                patches_mos, evaluation_image, stride_length=stride_length_vector), image.shape)
            mos_standard_deviation += pad_or_crop_image_to_size(reconstruct_image_from_patches(
                patches_mos_standard_deviation, evaluation_image, stride_length=stride_length_vector), image.shape)

        mos /= len(dimensions_to_predict)
        mos_standard_deviation /= len(dimensions_to_predict)

        if mask is None:

            if which_model == "tidsQualityAssessment":
                return_dict = {'MOS' : mos,
                               'MOS.standardDeviation' : mos_standard_deviation,
                               'MOS.mean' : mos.mean(),
                               'MOS.standardDeviationMean' : mos_standard_deviation.mean()
                              }
                return(return_dict)

            elif is_koniq:
                return_dict = {'MOS' : mos,
                               'sharpness' : mos_standard_deviation,
                               'MOS.mean' : mos.mean(),
                               'sharpness.mean' : mos_standard_deviation.mean()
                              }
                return(return_dict)

        else:

            if which_model == "tidsQualityAssessment":
                return_dict = {'MOS' : mos * mask,
                               'MOS.standardDeviation' : mos_standard_deviation * mask,
                               'MOS.mean' : (mos[mask >= 0.5]).mean(),
                               'MOS.standardDeviationMean' : (mos_standard_deviation[mask >= 0.5]).mean()
                              }
                return(return_dict)

            elif is_koniq:
                return_dict = {'MOS' : mos * mask,
                               'sharpness' : mos_standard_deviation * mask,
                               'MOS.mean' : (mos[mask >= 0.5]).mean(),
                               'sharpness.mean' : (mos_standard_deviation[mask >= 0.5]).mean()
                              }
                return(return_dict)
















import numpy as np
import tensorflow as tf
import ants


def mri_super_resolution(image, antsxnet_cache_directory=None, verbose=False):

    """
    Perform super-resolution (2x) of MRI data using deep back projection network.

    Arguments
    ---------
    image : ANTsImage
        magnetic resonance image

    antsxnet_cache_directory : string
        Destination directory for storing the downloaded template and model weights.
        Since these can be resused, if is None, these data will be downloaded to a
        ~/.keras/ANTsXNet/.

    verbose : boolean
        Print progress to the screen.

    Returns
    -------
    The super-resolved image.

    Example
    -------
    >>> image = ants.image_read("t1.nii.gz")
    >>> image_sr = mri_super_resolution(image)
    """

    from ..utilities import get_pretrained_network
    from ..utilities import apply_super_resolution_model_to_image
    from ..utilities import regression_match_image

    if image.dimension != 3:
        raise ValueError("Image dimension must be 3.")

    if antsxnet_cache_directory == None:
        antsxnet_cache_directory = "ANTsXNet"

    model_and_weights_file_name = get_pretrained_network("mriSuperResolution", antsxnet_cache_directory=antsxnet_cache_directory)
    model_sr = tf.keras.models.load_model(model_and_weights_file_name, compile=False)

    image_sr = apply_super_resolution_model_to_image(
        image, model_sr, target_range=(-127.5, 127.5)
    )
    image_sr = regression_match_image(
        image_sr, ants.resample_image_to_target(image, image_sr), poly_order=1
    )

    return image_sr



import tensorflow as tf

from tensorflow.keras.layers import Layer, InputSpec
from tensorflow.keras import initializers, regularizers, constraints
from tensorflow.keras import backend as K


class InstanceNormalization(Layer):

    """
    Instance normalization layer.

    Normalize the activations of the previous layer at each step,
    i.e. applies a transformation that maintains the mean activation
    close to 0 and the activation standard deviation close to 1.

    Taken from

    https://github.com/keras-team/keras-contrib/blob/master/keras_contrib/layers/normalization/instancenormalization.py

    Arguments
    ---------

    axis: integer
        Integer specifying which axis should be normalized, typically
        the feature axis.  For example, after a Conv2D layer with
        `channels_first`, set axis = 1.  Setting `axis=-1L` will
        normalize all values in each instance of the batch.  Axis 0
        is the batch dimension for tensorflow backend so we throw an
        error if `axis = 0`.

    epsilon: float
        Small float added to variance to avoid dividing by zero.

    center: If True, add offset of `beta` to normalized tensor.
        If False, `beta` is ignored.

    scale: If True, multiply by `gamma`.
        If False, `gamma` is not used.  When the next layer is linear (also e.g.,
        `nn.relu`), this can be disabled since the scaling will be done by the
        next layer.

    beta_initializer : string
        Initializer for the beta weight.

    gamma_initializer : string
        Initializer for the gamma weight.

    beta_regularizer : string
        Optional regularizer for the beta weight.

    gamma_regularizer : string
        Optional regularizer for the gamma weight.

    beta_constraint : string
        Optional constraint for the beta weight.

    gamma_constraint : string
        Optional constraint for the gamma weight.

    Returns
    -------
    Keras layer


    """
    def __init__(self, axis=None, epsilon=1e-3, center=True, scale=True,
                 beta_initializer='zeros', gamma_initializer='ones',
                 beta_regularizer=None, gamma_regularizer=None,
                 beta_constraint=None, gamma_constraint=None, **kwargs):

        super(InstanceNormalization, self).__init__(**kwargs)
        self.supports_masking = True
        self.axis = axis

        if self.axis == 0:
            raise ValueError('Axis cannot be zero')

        self.epsilon = epsilon
        self.center = center
        self.scale = scale
        self.beta_initializer = initializers.get(beta_initializer)
        self.gamma_initializer = initializers.get(gamma_initializer)
        self.beta_regularizer = regularizers.get(beta_regularizer)
        self.gamma_regularizer = regularizers.get(gamma_regularizer)
        self.beta_constraint = constraints.get(beta_constraint)
        self.gamma_constraint = constraints.get(gamma_constraint)

    def build(self, input_shape):
        dimensionality = len(input_shape)

        if (self.axis is not None) and (dimensionality == 2):
            raise ValueError('Cannot specify axis for rank 1 tensor')

        self.input_spec = InputSpec(ndim=dimensionality)

        if self.axis is None:
            shape = (1,)
        else:
            shape = (input_shape[self.axis],)

        if self.scale:
            self.gamma = self.add_weight(shape=shape,
                                         name='gamma',
                                         initializer=self.gamma_initializer,
                                         regularizer=self.gamma_regularizer,
                                         constraint=self.gamma_constraint)
        else:
            self.gamma = None

        if self.center:
            self.beta = self.add_weight(shape=shape,
                                        name='beta',
                                        initializer=self.beta_initializer,
                                        regularizer=self.beta_regularizer,
                                        constraint=self.beta_constraint)
        else:
            self.beta = None

        self.built = True

    def call(self, inputs, training=None):
        input_shape = K.int_shape(inputs)
        reduction_axes = list(range(0, len(input_shape)))

        if self.axis is not None:
            del reduction_axes[self.axis]

        del reduction_axes[0]

        mean = K.mean(inputs, reduction_axes, keepdims=True)
        stddev = K.std(inputs, reduction_axes, keepdims=True)
        normed = (inputs - mean) / (stddev + self.epsilon)

        broadcast_shape = [1] * len(input_shape)
        if self.axis is not None:
            broadcast_shape[self.axis] = input_shape[self.axis]

        if self.scale:
            broadcast_gamma = K.reshape(self.gamma, broadcast_shape)
            normed = normed * broadcast_gamma
        if self.center:
            broadcast_beta = K.reshape(self.beta, broadcast_shape)
            normed = normed + broadcast_beta
        return normed

    def get_config(self):
        config = {
            'axis': self.axis,
            'epsilon': self.epsilon,
            'center': self.center,
            'scale': self.scale,
            'beta_initializer': initializers.serialize(self.beta_initializer),
            'gamma_initializer': initializers.serialize(self.gamma_initializer),
            'beta_regularizer': regularizers.serialize(self.beta_regularizer),
            'gamma_regularizer': regularizers.serialize(self.gamma_regularizer),
            'beta_constraint': constraints.serialize(self.beta_constraint),
            'gamma_constraint': constraints.serialize(self.gamma_constraint)
        }
        base_config = super(InstanceNormalization, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


from tensorflow.keras.models import Model
from tensorflow.keras.layers import (Add, Activation, AveragePooling3D, BatchNormalization,
                          Conv3D, Dropout, Input, MaxPooling3D, ReLU, ZeroPadding3D)
from antspynet.utilities import LogSoftmax                          

def create_simple_fully_convolutional_network_model_3d(input_image_size,
                                                       number_of_filters_per_layer=(32, 64, 128, 256, 256, 64),
                                                       number_of_bins=40,
                                                       dropout_rate=0.5):
    """
    Implementation of the "SCFN" architecture for Brain/Gender prediction

    Creates a keras model implementation of the Simple Fully Convolutional
    Network model from the FMRIB group:

       https://github.com/ha-ha-ha-han/UKBiobank_deep_pretrain


    Arguments
    ---------
    input_image_size : tuple of length 4
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).
    number_of_filters_per_layer : array 
        number of filters for the convolutional layers.
    number_of_bins : integer
        number of bins for final softmax output.
    dropout_rate : float between 0 and 1
        Optional dropout rate before final convolution layer. 

    Returns
    -------
    Keras model
        A 3-D keras model.

    Example
    -------
    >>> model = create_simple_fully_convolutional_network_model_3d((None, None, None, 1))
    >>> model.summary()
    """

    number_of_layers = len(number_of_filters_per_layer)

    inputs = Input(shape=input_image_size)
 
    outputs = inputs
    for i in range(number_of_layers):
        if i < number_of_layers - 1:
            outputs = Conv3D(filters=number_of_filters_per_layer[i],
                             kernel_size=(3, 3, 3),
                             padding='valid')(outputs)
            outputs = ZeroPadding3D(padding=(1, 1, 1))(outputs)                 
            outputs = BatchNormalization(momentum=0.1,
                                         epsilon=1e-5)(outputs)
            outputs = MaxPooling3D(pool_size=(2, 2, 2),
                                   strides=(2, 2, 2))(outputs)
        else:
            outputs = Conv3D(filters=number_of_filters_per_layer[i],
                             kernel_size=(1, 1, 1),
                             padding='valid')(outputs)
            outputs = BatchNormalization(momentum=0.1,
                                         epsilon=1e-5)(outputs)
        outputs = ReLU()(outputs)

    outputs = AveragePooling3D(pool_size=(5, 6, 5),
                               strides=(5, 6, 5))(outputs)

    if dropout_rate > 0.0:
        outputs = Dropout(rate=dropout_rate)(outputs)

    outputs = Conv3D(filters=number_of_bins,
                     kernel_size=(1, 1, 1),
                     padding='valid')(outputs)
    outputs = LogSoftmax()(outputs)

    model = Model(inputs=inputs, outputs=outputs)

    return model




import statistics
import numpy as np
import tensorflow.keras as keras

import ants

def brain_age(t1,
              do_preprocessing=True,
              number_of_simulations=0,
              sd_affine=0.01,
              antsxnet_cache_directory=None,
              verbose=False):

    """
    Estimate BrainAge from a T1-weighted MR image using the DeepBrainNet
    architecture and weights described here:

    https://github.com/vishnubashyam/DeepBrainNet

    and described in the following article:

    https://academic.oup.com/brain/article-abstract/doi/10.1093/brain/awaa160/5863667?redirectedFrom=fulltext

    Preprocessing on the training data consisted of:
       * n4 bias correction,
       * brain extraction, and
       * affine registration to MNI.
    The input T1 should undergo the same steps.  If the input T1 is the raw
    T1, these steps can be performed by the internal preprocessing, i.e. set
    do_preprocessing = True

    Arguments
    ---------
    t1 : ANTsImage
        raw or preprocessed 3-D T1-weighted brain image.

    do_preprocessing : boolean
        See description above.

    number_of_simulations : integer
        Number of random affine perturbations to transform the input.

    sd_affine : float
        Define the standard deviation of the affine transformation parameter.

    antsxnet_cache_directory : string
        Destination directory for storing the downloaded template and model weights.
        Since these can be resused, if is None, these data will be downloaded to a
        ~/.keras/ANTsXNet/.

    verbose : boolean
        Print progress to the screen.

    Returns
    -------
    List consisting of the segmentation image and probability images for
    each label.

    Example
    -------
    >>> image = ants.image_read("t1.nii.gz")
    >>> deep = brain_age(image)
    >>> print("Predicted age: ", deep['predicted_age']
    """

    from ..utilities import preprocess_brain_image
    from ..utilities import get_pretrained_network
    from ..utilities import randomly_transform_image_data

    if t1.dimension != 3:
        raise ValueError( "Image dimension must be 3." )

    if antsxnet_cache_directory == None:
        antsxnet_cache_directory = "ANTsXNet"

    ################################
    #
    # Preprocess images
    #
    ################################

    t1_preprocessed = t1
    if do_preprocessing == True:
        t1_preprocessing = preprocess_brain_image(t1,
            truncate_intensity=(0.01, 0.99),
            brain_extraction_modality="t1",
            template="croppedMni152",
            template_transform_type="antsRegistrationSyNQuickRepro[a]",
            do_bias_correction=True,
            do_denoising=True,
            antsxnet_cache_directory=antsxnet_cache_directory,
            verbose=verbose)
        t1_preprocessed = t1_preprocessing["preprocessed_image"] * t1_preprocessing['brain_mask']

    t1_preprocessed = (t1_preprocessed - t1_preprocessed.min()) / (t1_preprocessed.max() - t1_preprocessed.min())

    ################################
    #
    # Load model and weights
    #
    ################################

    model_weights_file_name = get_pretrained_network("brainAgeDeepBrainNet", antsxnet_cache_directory=antsxnet_cache_directory)
    model = keras.models.load_model(model_weights_file_name)

    # The paper only specifies that 80 slices are used for prediction.  I just picked
    # a reasonable range spanning the center of the brain

    which_slices = list(range(45, 125))

    batchX = np.zeros((len(which_slices), *t1_preprocessed.shape[0:2], 3))

    input_image = list()
    input_image.append(t1_preprocessed)

    input_image_list = list()
    input_image_list.append(input_image)

    if number_of_simulations > 0:
        data_augmentation = randomly_transform_image_data(
            reference_image=t1_preprocessed,
            input_image_list=input_image_list,
            number_of_simulations=number_of_simulations,
            transform_type='affine',
            sd_affine=sd_affine,
            input_image_interpolator='linear')

    brain_age_per_slice = None
    for i in range(number_of_simulations + 1):

        batch_image = t1_preprocessed
        if i > 0:
            batch_image = data_augmentation['simulated_images'][i-1][0]

        for j in range(len(which_slices)):

            slice = (ants.slice_image(batch_image, axis=2, idx=which_slices[j])).numpy()
            batchX[j,:,:,0] = slice
            batchX[j,:,:,1] = slice
            batchX[j,:,:,2] = slice

        if verbose == True:
            print("Brain age (DeepBrainNet):  predicting brain age per slice (batch = ", i, ")")

        if i == 0:
            brain_age_per_slice = model.predict(batchX, verbose=verbose)
        else:
            prediction = model.predict(batchX, verbose=verbose)
            brain_age_per_slice = brain_age_per_slice + (prediction - brain_age_per_slice) /  (i+1)

    predicted_age = statistics.median(brain_age_per_slice)[0]

    return_dict = {'predicted_age' : predicted_age,
                   'brain_age_per_slice' : brain_age_per_slice}
    return(return_dict)



import tensorflow as tf

def get_antsxnet_data(file_id=None,
                      target_file_name=None,
                      antsxnet_cache_directory=None):

    """
    Download data such as prefabricated templates and spatial priors.

    Arguments
    ---------

    file_id string
        One of the permitted file ids or pass "show" to list all
        valid possibilities. Note that most require internet access
        to download.

    target_file_name string
       Optional target filename.

    antsxnet_cache_directory string
       Optional target output.  If not specified these data will be downloaded
       to the subdirectory ~/.keras/ANTsXNet/.

    Returns
    -------
    A filename string

    Example
    -------
    >>> template_file = get_antsxnet_data('biobank')
    """

    def switch_networks(argument):
        switcher = {
            "biobank": "https://ndownloader.figshare.com/files/22429242",
            "croppedMni152": "https://ndownloader.figshare.com/files/22933754",
            "croppedMni152Priors": "https://ndownloader.figshare.com/files/27688437",
            "mprage_hippmapp3r": "https://ndownloader.figshare.com/files/24984689",
            "protonLungTemplate": "https://ndownloader.figshare.com/files/22707338",
            "ctLungTemplate": "https://ndownloader.figshare.com/files/22707335",
            "luna16LungPriors": "https://ndownloader.figshare.com/files/28253796",
            "priorDktLabels": "https://ndownloader.figshare.com/files/24139802",
            "S_template3": "https://ndownloader.figshare.com/files/22597175",
            "priorDeepFlashLeftLabels": "https://ndownloader.figshare.com/files/25422098",
            "priorDeepFlashRightLabels": "https://ndownloader.figshare.com/files/25422101",
            "adni": "https://ndownloader.figshare.com/files/25516361",
            "ixi": "https://ndownloader.figshare.com/files/25516358",
            "kirby": "https://ndownloader.figshare.com/files/25620107",
            "mni152": "https://ndownloader.figshare.com/files/25516349",
            "nki": "https://ndownloader.figshare.com/files/25516355",
            "nki10": "https://ndownloader.figshare.com/files/25516346",
            "oasis": "https://ndownloader.figshare.com/files/25516352"
        }
        return(switcher.get(argument, "Invalid argument."))

    if file_id == None:
        raise ValueError("Missing file id.")

    valid_list = ("biobank",
                  "croppedMni152",
                  "croppedMni152Priors",
                  "mprage_hippmapp3r",
                  "protonLungTemplate",
                  "ctLungTemplate",
                  "luna16LungPriors",
                  "S_template3",
                  "priorDktLabels",
                  "priorDeepFlashLeftLabels",
                  "priorDeepFlashRightLabels",
                  "adni",
                  "ixi",
                  "kirby",
                  "mni152",
                  "nki",
                  "nki10",
                  "oasis",
                  "show")

    if not file_id in valid_list:
        raise ValueError("No data with the id you passed - try \"show\" to get list of valid ids.")

    if file_id == "show":
       return(valid_list)

    url = switch_networks(file_id)

    if target_file_name == None:
        target_file_name = file_id + ".nii.gz"

    if antsxnet_cache_directory == None:
        antsxnet_cache_directory = "ANTsXNet"

    target_file_name_path = tf.keras.utils.get_file(target_file_name, url,
        cache_subdir=antsxnet_cache_directory)

    return(target_file_name_path)




import numpy as np
import ants

def el_bicho(ventilation_image,
             mask,
             use_coarse_slices_only=True,
             antsxnet_cache_directory=None,
             verbose=False):

    """
    Perform functional lung segmentation using hyperpolarized gases.

    https://pubmed.ncbi.nlm.nih.gov/30195415/

    Arguments
    ---------
    ventilation_image : ANTsImage
        input ventilation image.

    mask : ANTsImage
        input mask.

    use_coarse_slices_only : boolean
        If True, apply network only in the dimension of greatest slice thickness.
        If False, apply to all dimensions and average the results.

    antsxnet_cache_directory : string
        Destination directory for storing the downloaded template and model weights.
        Since these can be resused, if is None, these data will be downloaded to a
        ~/.keras/ANTsXNet/.

    verbose : boolean
        Print progress to the screen.

    Returns
    -------
    Ventilation segmentation and corresponding probability images

    Example
    -------
    >>> image = ants.image_read("ventilation.nii.gz")
    >>> mask = ants.image_read("mask.nii.gz")
    >>> lung_seg = el_bicho(image, mask, use_coarse_slices=True, verbose=False)
    """

    from ..architectures import create_unet_model_2d
    from ..utilities import get_pretrained_network
    from ..utilities import pad_or_crop_image_to_size

    if ventilation_image.dimension != 3:
        raise ValueError("Image dimension must be 3.")

    if ventilation_image.shape != mask.shape:
        raise ValueError("Ventilation image and mask size are not the same size.")

    if antsxnet_cache_directory == None:
        antsxnet_cache_directory = "ANTsXNet"

    ################################
    #
    # Preprocess image
    #
    ################################

    template_size = (256, 256)
    classes = (0, 1, 2, 3, 4)
    number_of_classification_labels = len(classes)

    image_modalities = ("Ventilation", "Mask")
    channel_size = len(image_modalities)

    preprocessed_image = (ventilation_image - ventilation_image.mean()) / ventilation_image.std()
    ants.set_direction(preprocessed_image, np.identity(3))

    mask_identity = ants.image_clone(mask)
    ants.set_direction(mask_identity, np.identity(3))

    ################################
    #
    # Build models and load weights
    #
    ################################

    unet_model = create_unet_model_2d((*template_size, channel_size),
        number_of_outputs=number_of_classification_labels,
        number_of_layers=4, number_of_filters_at_base_layer=32, dropout_rate=0.0,
        convolution_kernel_size=(3, 3), deconvolution_kernel_size=(2, 2),
        weight_decay=1e-5, additional_options=("attentionGating"))

    if verbose == True:
        print("El Bicho: retrieving model weights.")

    weights_file_name = get_pretrained_network("elBicho", antsxnet_cache_directory=antsxnet_cache_directory)
    unet_model.load_weights(weights_file_name)

    ################################
    #
    # Extract slices
    #
    ################################

    spacing = ants.get_spacing(preprocessed_image)
    dimensions_to_predict = (spacing.index(max(spacing)),)
    if use_coarse_slices_only == False:
        dimensions_to_predict = list(range(3))

    total_number_of_slices = 0
    for d in range(len(dimensions_to_predict)):
        total_number_of_slices += preprocessed_image.shape[dimensions_to_predict[d]]

    batchX = np.zeros((total_number_of_slices, *template_size, channel_size))

    slice_count = 0
    for d in range(len(dimensions_to_predict)):
        number_of_slices = preprocessed_image.shape[dimensions_to_predict[d]]

        if verbose == True:
            print("Extracting slices for dimension ", dimensions_to_predict[d], ".")

        for i in range(number_of_slices):
            ventilation_slice = pad_or_crop_image_to_size(ants.slice_image(preprocessed_image, dimensions_to_predict[d], i), template_size)
            batchX[slice_count,:,:,0] = ventilation_slice.numpy()

            mask_slice = pad_or_crop_image_to_size(ants.slice_image(mask_identity, dimensions_to_predict[d], i), template_size)
            batchX[slice_count,:,:,1] = mask_slice.numpy()

            slice_count += 1


    ################################
    #
    # Do prediction and then restack into the image
    #
    ################################

    if verbose == True:
        print("Prediction.")

    prediction = unet_model.predict(batchX, verbose=verbose)

    permutations = list()
    permutations.append((0, 1, 2))
    permutations.append((1, 0, 2))
    permutations.append((1, 2, 0))

    probability_images = list()
    for l in range(number_of_classification_labels):
        probability_images.append(ants.image_clone(mask) * 0)

    current_start_slice = 0
    for d in range(len(dimensions_to_predict)):
        current_end_slice = current_start_slice + preprocessed_image.shape[dimensions_to_predict[d]] - 1
        which_batch_slices = range(current_start_slice, current_end_slice)

        for l in range(number_of_classification_labels):
            prediction_per_dimension = prediction[which_batch_slices,:,:,l]
            prediction_array = np.transpose(np.squeeze(prediction_per_dimension), permutations[dimensions_to_predict[d]])
            prediction_image = ants.copy_image_info(ventilation_image,
                pad_or_crop_image_to_size(ants.from_numpy(prediction_array),
                ventilation_image.shape))
            probability_images[l] = probability_images[l] + (prediction_image - probability_images[l]) / (d + 1)

        current_start_slice = current_end_slice + 1

    ################################
    #
    # Convert probability images to segmentation
    #
    ################################

    image_matrix = ants.image_list_to_matrix(probability_images[1:(len(probability_images))], mask * 0 + 1)
    background_foreground_matrix = np.stack([ants.image_list_to_matrix([probability_images[0]], mask * 0 + 1),
                                            np.expand_dims(np.sum(image_matrix, axis=0), axis=0)])
    foreground_matrix = np.argmax(background_foreground_matrix, axis=0)
    segmentation_matrix = (np.argmax(image_matrix, axis=0) + 1) * foreground_matrix
    segmentation_image = ants.matrix_to_images(
        np.expand_dims(segmentation_matrix, axis=0), mask * 0 + 1)[0]

    return_dict = {'segmentation_image' : segmentation_image,
                   'probability_images' : probability_images}
    return(return_dict)





import numpy as np
import ants

def claustrum_segmentation(t1,
                           do_preprocessing=True,
                           use_ensemble=True,
                           antsxnet_cache_directory=None,
                           verbose=False):

    """
    Claustrum segmentation

    Described here:

        https://arxiv.org/abs/2008.03465

    with the implementation available at:

        https://github.com/hongweilibran/claustrum_multi_view


    Arguments
    ---------
    t1 : ANTsImage
        input 3-D T1 brain image.

    do_preprocessing : boolean
        perform n4 bias correction.

    use_ensemble : boolean
        check whether to use all 3 sets of weights.

    antsxnet_cache_directory : string
        Destination directory for storing the downloaded template and model weights.
        Since these can be resused, if is None, these data will be downloaded to a
        ~/.keras/ANTsXNet/.

    verbose : boolean
        Print progress to the screen.

    Returns
    -------
    Claustrum segmentation probability image

    Example
    -------
    >>> image = ants.image_read("t1.nii.gz")
    >>> probability_mask = claustrum_segmentation(image)
    """

    from ..architectures import create_sysu_media_unet_model_2d
    from ..utilities import brain_extraction
    from ..utilities import get_pretrained_network
    from ..utilities import preprocess_brain_image
    from ..utilities import pad_or_crop_image_to_size

    if t1.dimension != 3:
        raise ValueError( "Image dimension must be 3." )

    if antsxnet_cache_directory == None:
        antsxnet_cache_directory = "ANTsXNet"

    image_size = (180, 180)

    ################################
    #
    # Preprocess images
    #
    ################################

    number_of_channels = 1
    t1_preprocessed = ants.image_clone(t1)
    brain_mask = ants.threshold_image(t1, 0, 0, 0, 1)
    if do_preprocessing == True:
        t1_preprocessing = preprocess_brain_image(t1,
            truncate_intensity=(0.01, 0.99),
            brain_extraction_modality="t1",
            do_bias_correction=True,
            do_denoising=True,
            antsxnet_cache_directory=antsxnet_cache_directory,
            verbose=verbose)
        t1_preprocessed = t1_preprocessing["preprocessed_image"]
        brain_mask = t1_preprocessing["brain_mask"]

    reference_image = ants.make_image((170, 256, 256),
                                       voxval=1,
                                       spacing=(1, 1, 1),
                                       origin=(0, 0, 0),
                                       direction=np.identity(3))
    center_of_mass_reference = ants.get_center_of_mass(reference_image)
    center_of_mass_image = ants.get_center_of_mass(brain_mask)
    translation = np.asarray(center_of_mass_image) - np.asarray(center_of_mass_reference)
    xfrm = ants.create_ants_transform(transform_type="Euler3DTransform",
        center=np.asarray(center_of_mass_reference), translation=translation)
    t1_preprocessed_warped = ants.apply_ants_transform_to_image(xfrm, t1_preprocessed, reference_image)
    brain_mask_warped = ants.threshold_image(
        ants.apply_ants_transform_to_image(xfrm, brain_mask, reference_image), 0.5, 1.1, 1, 0 )

    ################################
    #
    # Gaussian normalize intensity based on brain mask
    #
    ################################

    mean_t1 = t1_preprocessed_warped[brain_mask_warped > 0].mean()
    std_t1 = t1_preprocessed_warped[brain_mask_warped > 0].std()
    t1_preprocessed_warped = (t1_preprocessed_warped - mean_t1) / std_t1

    t1_preprocessed_warped = t1_preprocessed_warped * brain_mask_warped

    ################################
    #
    # Build models and load weights
    #
    ################################

    number_of_models = 1
    if use_ensemble == True:
        number_of_models = 3

    if verbose == True:
        print("Claustrum:  retrieving axial model weights.")

    unet_axial_models = list()
    for i in range(number_of_models):
        weights_file_name = get_pretrained_network("claustrum_axial_" + str(i), antsxnet_cache_directory=antsxnet_cache_directory)
        unet_axial_models.append(create_sysu_media_unet_model_2d((*image_size, number_of_channels), anatomy="claustrum"))
        unet_axial_models[i].load_weights(weights_file_name)

    if verbose == True:
        print("Claustrum:  retrieving coronal model weights.")

    unet_coronal_models = list()
    for i in range(number_of_models):
        weights_file_name = get_pretrained_network("claustrum_coronal_" + str(i), antsxnet_cache_directory=antsxnet_cache_directory)
        unet_coronal_models.append(create_sysu_media_unet_model_2d((*image_size, number_of_channels), anatomy="claustrum"))
        unet_coronal_models[i].load_weights(weights_file_name)

    ################################
    #
    # Extract slices
    #
    ################################

    dimensions_to_predict = [1, 2]

    batch_coronal_X = np.zeros((t1_preprocessed_warped.shape[1], *image_size, number_of_channels))
    batch_axial_X = np.zeros((t1_preprocessed_warped.shape[2], *image_size, number_of_channels))

    for d in range(len(dimensions_to_predict)):
        number_of_slices = t1_preprocessed_warped.shape[dimensions_to_predict[d]]

        if verbose == True:
            print("Extracting slices for dimension ", dimensions_to_predict[d], ".")

        for i in range(number_of_slices):
            t1_slice = pad_or_crop_image_to_size(ants.slice_image(t1_preprocessed_warped, dimensions_to_predict[d], i), image_size)
            if dimensions_to_predict[d] == 1:
                batch_coronal_X[i,:,:,0] = np.rot90(t1_slice.numpy(), k=-1)
            else:
                batch_axial_X[i,:,:,0] = np.rot90(t1_slice.numpy())

    ################################
    #
    # Do prediction and then restack into the image
    #
    ################################

    if verbose == True:
        print("Coronal prediction.")

    prediction_coronal = unet_coronal_models[0].predict(batch_coronal_X, verbose=verbose)
    if number_of_models > 1:
       for i in range(1, number_of_models, 1):
           prediction_coronal += unet_coronal_models[i].predict(batch_coronal_X, verbose=verbose)
    prediction_coronal /= number_of_models

    for i in range(t1_preprocessed_warped.shape[1]):
        prediction_coronal[i,:,:,0] = np.rot90(np.squeeze(prediction_coronal[i,:,:,0]))

    if verbose == True:
        print("Axial prediction.")

    prediction_axial = unet_axial_models[0].predict(batch_axial_X, verbose=verbose)
    if number_of_models > 1:
       for i in range(1, number_of_models, 1):
           prediction_axial += unet_axial_models[i].predict(batch_axial_X, verbose=verbose)
    prediction_axial /= number_of_models

    for i in range(t1_preprocessed_warped.shape[2]):
        prediction_axial[i,:,:,0] = np.rot90(np.squeeze(prediction_axial[i,:,:,0]), k=-1)

    if verbose == True:
        print("Restack image and transform back to native space.")

    permutations = list()
    permutations.append((0, 1, 2))
    permutations.append((1, 0, 2))
    permutations.append((1, 2, 0))

    prediction_image_average = ants.image_clone(t1_preprocessed_warped) * 0

    for d in range(len(dimensions_to_predict)):
        which_batch_slices = range(t1_preprocessed_warped.shape[dimensions_to_predict[d]])
        prediction_per_dimension = None
        if dimensions_to_predict[d] == 1:
            prediction_per_dimension = prediction_coronal[which_batch_slices,:,:,:]
        else:
            prediction_per_dimension = prediction_axial[which_batch_slices,:,:,:]
        prediction_array = np.transpose(np.squeeze(prediction_per_dimension), permutations[dimensions_to_predict[d]])
        prediction_image = ants.copy_image_info(t1_preprocessed_warped,
          pad_or_crop_image_to_size(ants.from_numpy(prediction_array),
            t1_preprocessed_warped.shape))
        prediction_image_average = prediction_image_average + (prediction_image - prediction_image_average) / (d + 1)

    probability_image = ants.apply_ants_transform_to_image(ants.invert_ants_transform(xfrm),
        prediction_image_average, t1) * ants.threshold_image(brain_mask, 0.5, 1, 1, 0)

    return(probability_image)




import ants
import numpy as np
from tensorflow import keras

def sysu_media_wmh_segmentation(flair,
                                t1=None,
                                use_ensemble=True,
                                antsxnet_cache_directory=None,
                                verbose=False):

    """
    Perform WMH segmentation using the winning submission in the MICCAI
    2017 challenge by the sysu_media team using FLAIR or T1/FLAIR.  The
    MICCAI challenge is discussed in

    https://pubmed.ncbi.nlm.nih.gov/30908194/

    with the sysu_media's team entry is discussed in

     https://pubmed.ncbi.nlm.nih.gov/30125711/

    with the original implementation available here:

    https://github.com/hongweilibran/wmh_ibbmTum

    Arguments
    ---------
    flair : ANTsImage
        input 3-D FLAIR brain image (not skull-stripped).

    t1 : ANTsImage
        input 3-D T1 brain image (not skull-stripped).

    use_ensemble : boolean
        check whether to use all 3 sets of weights.

    antsxnet_cache_directory : string
        Destination directory for storing the downloaded template and model weights.
        Since these can be resused, if is None, these data will be downloaded to a
        ~/.keras/ANTsXNet/.

    verbose : boolean
        Print progress to the screen.

    Returns
    -------
    WMH segmentation probability image

    Example
    -------
    >>> image = ants.image_read("flair.nii.gz")
    >>> probability_mask = sysu_media_wmh_segmentation(image)
    """

    from ..architectures import create_sysu_media_unet_model_2d
    from ..utilities import get_pretrained_network
    from ..utilities import pad_or_crop_image_to_size
    from ..utilities import preprocess_brain_image
    from ..utilities import binary_dice_coefficient

    if flair.dimension != 3:
        raise ValueError( "Image dimension must be 3." )

    if antsxnet_cache_directory == None:
        antsxnet_cache_directory = "ANTsXNet"

    image_size = (200, 200)

    ################################
    #
    # Preprocess images
    #
    ################################

    def closest_simplified_direction_matrix(direction):
        closest = np.floor(np.abs(direction + 0.5))
        closest[direction < 0] *= -1.0
        return direction

    simplified_direction = closest_simplified_direction_matrix(flair.direction)  

    flair_preprocessing = preprocess_brain_image(flair,
        truncate_intensity=None,
        brain_extraction_modality=None,
        do_bias_correction=False,
        do_denoising=False,
        antsxnet_cache_directory=antsxnet_cache_directory,
        verbose=verbose)
    flair_preprocessed = flair_preprocessing["preprocessed_image"]
    flair_preprocessed.set_direction(simplified_direction)
    flair_preprocessed.set_origin((0, 0, 0))
    flair_preprocessed.set_spacing((1, 1, 1))
    number_of_channels = 1

    t1_preprocessed = None
    if t1 is not None:
        t1_preprocessing = preprocess_brain_image(t1,
            truncate_intensity=None,
            brain_extraction_modality=None,
            do_bias_correction=False,
            do_denoising=False,
            antsxnet_cache_directory=antsxnet_cache_directory,
            verbose=verbose)
        t1_preprocessed = t1_preprocessing["preprocessed_image"]
        t1_preprocessed.set_direction(simplified_direction)
        t1_preprocessed.set_origin((0, 0, 0))
        t1_preprocessed.set_spacing((1, 1, 1))
        number_of_channels = 2

    ################################
    #
    # Reorient images
    #
    ################################

    reference_image = ants.make_image((256, 256, 256),
                                       voxval=0,
                                       spacing=(1, 1, 1),
                                       origin=(0, 0, 0),
                                       direction=np.identity(3))
    center_of_mass_reference = np.floor(ants.get_center_of_mass(reference_image * 0 + 1))
    center_of_mass_image = np.floor(ants.get_center_of_mass(flair_preprocessed))
    translation = np.asarray(center_of_mass_image) - np.asarray(center_of_mass_reference)
    xfrm = ants.create_ants_transform(transform_type="Euler3DTransform",
        center=np.asarray(center_of_mass_reference), translation=translation)
    flair_preprocessed_warped = ants.apply_ants_transform_to_image(
        xfrm, flair_preprocessed, reference_image, interpolation="nearestneighbor")
    crop_image = ants.image_clone(flair_preprocessed) * 0 + 1
    crop_image_warped = ants.apply_ants_transform_to_image(
        xfrm, crop_image, reference_image, interpolation="nearestneighbor")
    flair_preprocessed_warped = ants.crop_image(flair_preprocessed_warped, crop_image_warped, 1)

    if t1 is not None:
        t1_preprocessed_warped = ants.apply_ants_transform_to_image(
            xfrm, t1_preprocessed, reference_image, interpolation="nearestneighbor")
        t1_preprocessed_warped = ants.crop_image(t1_preprocessed_warped, crop_image_warped, 1) 

    ################################
    #
    # Gaussian normalize intensity
    #
    ################################

    mean_flair = flair_preprocessed.mean()
    std_flair = flair_preprocessed.std()
    if number_of_channels == 2:
        mean_t1 = t1_preprocessed.mean()
        std_t1 = t1_preprocessed.std()

    flair_preprocessed_warped = (flair_preprocessed_warped - mean_flair) / std_flair
    if number_of_channels == 2:
        t1_preprocessed_warped = (t1_preprocessed_warped - mean_t1) / std_t1

    ################################
    #
    # Build models and load weights
    #
    ################################

    number_of_models = 1
    if use_ensemble == True:
        number_of_models = 3

    if verbose == True:
        print("White matter hyperintensity:  retrieving model weights.")

    unet_models = list()
    for i in range(number_of_models):
        if number_of_channels == 1:
            weights_file_name = get_pretrained_network("sysuMediaWmhFlairOnlyModel" + str(i), 
                antsxnet_cache_directory=antsxnet_cache_directory)
        else:
            weights_file_name = get_pretrained_network("sysuMediaWmhFlairT1Model" + str(i), 
                antsxnet_cache_directory=antsxnet_cache_directory)
        unet_model = create_sysu_media_unet_model_2d((*image_size, number_of_channels))             
        unet_loss = binary_dice_coefficient(smoothing_factor=1.)
        unet_model.compile(optimizer=keras.optimizers.Adam(learning_rate=2e-4),
                        loss=unet_loss)
        unet_model.load_weights(weights_file_name)                
        unet_models.append(unet_model)
        

    ################################
    #
    # Extract slices
    #
    ################################

    dimensions_to_predict = [2]

    total_number_of_slices = 0
    for d in range(len(dimensions_to_predict)):
        total_number_of_slices += flair_preprocessed_warped.shape[dimensions_to_predict[d]]

    batchX = np.zeros((total_number_of_slices, *image_size, number_of_channels))

    slice_count = 0
    for d in range(len(dimensions_to_predict)):
        number_of_slices = flair_preprocessed_warped.shape[dimensions_to_predict[d]]

        if verbose == True:
            print("Extracting slices for dimension ", dimensions_to_predict[d], ".")

        for i in range(number_of_slices):
            flair_slice = pad_or_crop_image_to_size(ants.slice_image(flair_preprocessed_warped, dimensions_to_predict[d], i), image_size)
            batchX[slice_count,:,:,0] = flair_slice.numpy()
            if number_of_channels == 2:
                t1_slice = pad_or_crop_image_to_size(ants.slice_image(t1_preprocessed_warped, dimensions_to_predict[d], i), image_size)
                batchX[slice_count,:,:,1] = t1_slice.numpy()
            slice_count += 1

    ################################
    #
    # Do prediction and then restack into the image
    #
    ################################

    if verbose == True:
        print("Prediction.")

    prediction = unet_models[0].predict(np.transpose(batchX, axes=(0, 2, 1, 3)), verbose=verbose)
    if number_of_models > 1:
        for i in range(1, number_of_models, 1):
            prediction += unet_models[i].predict(np.transpose(batchX, axes=(0, 2, 1, 3)), verbose=verbose)
    prediction /= number_of_models
    prediction = np.transpose(prediction, axes=(0, 2, 1, 3))

    permutations = list()
    permutations.append((0, 1, 2))
    permutations.append((1, 0, 2))
    permutations.append((1, 2, 0))

    prediction_image_average = ants.image_clone(flair_preprocessed_warped) * 0

    current_start_slice = 0
    for d in range(len(dimensions_to_predict)):
        current_end_slice = current_start_slice + flair_preprocessed_warped.shape[dimensions_to_predict[d]]
        which_batch_slices = range(current_start_slice, current_end_slice)
        prediction_per_dimension = prediction[which_batch_slices,:,:,:]
        prediction_array = np.transpose(np.squeeze(prediction_per_dimension), permutations[dimensions_to_predict[d]])
        prediction_image = ants.copy_image_info(flair_preprocessed_warped,
          pad_or_crop_image_to_size(ants.from_numpy(prediction_array),
            flair_preprocessed_warped.shape))
        prediction_image_average = prediction_image_average + (prediction_image - prediction_image_average) / (d + 1)
        current_start_slice = current_end_slice

    probability_image = ants.apply_ants_transform_to_image(
        ants.invert_ants_transform(xfrm), prediction_image_average, flair_preprocessed)
    probability_image = ants.copy_image_info(flair, probability_image)    

    return(probability_image)

def ew_david(flair,
             t1,
             do_preprocessing=True,
             which_model="sysu",
             which_axes=2,
             number_of_simulations=0,
             sd_affine=0.01,
             antsxnet_cache_directory=None,
             verbose=False):

    """
    Perform White matter hypterintensity probabilistic segmentation
    using deep learning

    Preprocessing on the training data consisted of:
       * n4 bias correction,
       * intensity truncation,
       * brain extraction, and
       * affine registration to MNI.
    The input T1 should undergo the same steps.  If the input T1 is the raw
    T1, these steps can be performed by the internal preprocessing, i.e. set
    \code{do_preprocessing = True}

    Arguments
    ---------
    flair : ANTsImage
        input 3-D FLAIR brain image (not skull-stripped).

    t1 : ANTsImage
        input 3-D T1 brain image (not skull-stripped).

    do_preprocessing : boolean
        perform n4 bias correction, intensity truncation, brain extraction.

    which_model : string
       one of "sysu", "sysuPlus", "sysuPlusSeg", "sysuWithSite".

    which_axes : string or scalar or tuple/vector
        apply 2-D model to 1 or more axes.  In addition to a scalar
        or vector, e.g., which_axes = (0, 2), one can use "max" for the
        axis with maximum anisotropy (default) or "all" for all axes.

    number_of_simulations : integer
        Number of random affine perturbations to transform the input.

    sd_affine : float
        Define the standard deviation of the affine transformation parameter.

    antsxnet_cache_directory : string
        Destination directory for storing the downloaded template and model weights.
        Since these can be resused, if is None, these data will be downloaded to a
        ~/.keras/ANTsXNet/.

    verbose : boolean
        Print progress to the screen.

    Returns
    -------
    WMH segmentation probability image

    Example
    -------
    >>> image = ants.image_read("flair.nii.gz")
    >>> probability_mask = sysu_media_wmh_segmentation(image)
    """

    from ..architectures import create_unet_model_2d
    from ..utilities import deep_atropos
    from ..utilities import get_pretrained_network
    from ..utilities import preprocess_brain_image
    from ..utilities import randomly_transform_image_data
    from ..utilities import pad_or_crop_image_to_size

    do_t1_only = False
    do_flair_only = False

    if flair is None and t1 is not None:
        do_t1_only = True
    elif flair is not None and t1 is None:
        do_flair_only = True

    use_t1_segmentation = False
    if "Seg" in which_model:
        if do_flair_only:
            raise ValueError("Segmentation requires T1.")
        else:
            use_t1_segmentation = True

    if use_t1_segmentation and do_preprocessing == False:
        raise ValueError("Using the t1 segmentation requires do_preprocessing=True.")

    if antsxnet_cache_directory == None:
        antsxnet_cache_directory = "ANTsXNet"

    do_slicewise = True

    if do_slicewise == False:

        raise ValueError("Not available.")

        # ################################
        # #
        # # Preprocess images
        # #
        # ################################

        # t1_preprocessed = t1
        # t1_preprocessing = None
        # if do_preprocessing == True:
        #     t1_preprocessing = preprocess_brain_image(t1,
        #         truncate_intensity=(0.01, 0.99),
        #         brain_extraction_modality="t1",
        #         template="croppedMni152",
        #         template_transform_type="antsRegistrationSyNQuickRepro[a]",
        #         do_bias_correction=True,
        #         do_denoising=False,
        #         antsxnet_cache_directory=antsxnet_cache_directory,
        #         verbose=verbose)
        #     t1_preprocessed = t1_preprocessing["preprocessed_image"] * t1_preprocessing['brain_mask']

        # flair_preprocessed = flair
        # if do_preprocessing == True:
        #     flair_preprocessing = preprocess_brain_image(flair,
        #         truncate_intensity=(0.01, 0.99),
        #         brain_extraction_modality="t1",
        #         do_bias_correction=True,
        #         do_denoising=False,
        #         antsxnet_cache_directory=antsxnet_cache_directory,
        #         verbose=verbose)
        #     flair_preprocessed = ants.apply_transforms(fixed=t1_preprocessed,
        #         moving=flair_preprocessing["preprocessed_image"],
        #         transformlist=t1_preprocessing['template_transforms']['fwdtransforms'])
        #     flair_preprocessed = flair_preprocessed * t1_preprocessing['brain_mask']

        # ################################
        # #
        # # Build model and load weights
        # #
        # ################################

        # patch_size = (112, 112, 112)
        # stride_length = (t1_preprocessed.shape[0] - patch_size[0],
        #                 t1_preprocessed.shape[1] - patch_size[1],
        #                 t1_preprocessed.shape[2] - patch_size[2])

        # classes = ("background", "wmh" )
        # number_of_classification_labels = len(classes)
        # labels = (0, 1)

        # image_modalities = ("T1", "FLAIR")
        # channel_size = len(image_modalities)

        # unet_model = create_unet_model_3d((*patch_size, channel_size),
        #     number_of_outputs = number_of_classification_labels,
        #     number_of_layers = 4, number_of_filters_at_base_layer = 16, dropout_rate = 0.0,
        #     convolution_kernel_size = (3, 3, 3), deconvolution_kernel_size = (2, 2, 2),
        #     weight_decay = 1e-5, additional_options=("attentionGating"))

        # weights_file_name = get_pretrained_network("ewDavidWmhSegmentationWeights",
        #     antsxnet_cache_directory=antsxnet_cache_directory)
        # unet_model.load_weights(weights_file_name)

        # ################################
        # #
        # # Do prediction and normalize to native space
        # #
        # ################################

        # if verbose == True:
        #     print("ew_david:  prediction.")

        # batchX = np.zeros((8, *patch_size, channel_size))

        # t1_preprocessed = (t1_preprocessed - t1_preprocessed.mean()) / t1_preprocessed.std()
        # t1_patches = extract_image_patches(t1_preprocessed, patch_size=patch_size,
        #                                     max_number_of_patches="all", stride_length=stride_length,
        #                                     return_as_array=True)
        # batchX[:,:,:,:,0] = t1_patches

        # flair_preprocessed = (flair_preprocessed - flair_preprocessed.mean()) / flair_preprocessed.std()
        # flair_patches = extract_image_patches(flair_preprocessed, patch_size=patch_size,
        #                                     max_number_of_patches="all", stride_length=stride_length,
        #                                     return_as_array=True)
        # batchX[:,:,:,:,1] = flair_patches

        # predicted_data = unet_model.predict(batchX, verbose=verbose)

        # probability_images = list()
        # for i in range(len(labels)):
        #     print("Reconstructing image", classes[i])
        #     reconstructed_image = reconstruct_image_from_patches(predicted_data[:,:,:,:,i],
        #         domain_image=t1_preprocessed, stride_length=stride_length)

        #     if do_preprocessing == True:
        #         probability_images.append(ants.apply_transforms(fixed=t1,
        #             moving=reconstructed_image,
        #             transformlist=t1_preprocessing['template_transforms']['invtransforms'],
        #             whichtoinvert=[True], interpolator="linear", verbose=verbose))
        #     else:
        #         probability_images.append(reconstructed_image)

        # return(probability_images[1])

    else:  # do_slicewise

        ################################
        #
        # Preprocess images
        #
        ################################

        t1_preprocessed = None
        t1_preprocessing = None
        brain_mask = None
        if t1 is not None:
            if do_preprocessing == True:
                t1_preprocessing = preprocess_brain_image(t1,
                    truncate_intensity=None,
                    brain_extraction_modality="t1",
                    do_bias_correction=False,
                    do_denoising=False,
                    antsxnet_cache_directory=antsxnet_cache_directory,
                    verbose=verbose)
                brain_mask = ants.threshold_image(t1_preprocessing["brain_mask"], 0.5, 1, 1, 0)
                t1_preprocessed = t1_preprocessing["preprocessed_image"] * brain_mask

        t1_segmentation = None
        if use_t1_segmentation:
            atropos_seg = deep_atropos(t1, do_preprocessing=True, verbose=verbose)
            t1_segmentation = atropos_seg['segmentation_image']

        flair_preprocessed = None
        if flair is not None:
            flair_preprocessed = flair
            if do_preprocessing == True:
                if brain_mask is None:  
                    flair_preprocessing = preprocess_brain_image(flair,
                        truncate_intensity=None,
                        brain_extraction_modality="flair",
                        do_bias_correction=False,
                        do_denoising=False,
                        antsxnet_cache_directory=antsxnet_cache_directory,
                        verbose=verbose)
                    brain_mask = flair_preprocessing["brain_mask"]
                else:
                    flair_preprocessing = preprocess_brain_image(flair,
                        truncate_intensity=None,
                        brain_extraction_modality=None,
                        do_bias_correction=False,
                        do_denoising=False,
                        antsxnet_cache_directory=antsxnet_cache_directory,
                        verbose=verbose)
                flair_preprocessed = flair_preprocessing["preprocessed_image"] * brain_mask

        if t1_preprocessed is not None:
            resampling_params = list(ants.get_spacing(t1_preprocessed))
        else:
            resampling_params = list(ants.get_spacing(flair_preprocessed))

        do_resampling = False
        for d in range(len(resampling_params)):
            if resampling_params[d] < 0.8:
                resampling_params[d] = 1.0
                do_resampling = True

        resampling_params = tuple(resampling_params)

        if do_resampling:
            if flair_preprocessed is not None:
                flair_preprocessed = ants.resample_image(flair_preprocessed, resampling_params, use_voxels=False, interp_type=0)
            if t1_preprocessed is not None:
                t1_preprocessed = ants.resample_image(t1_preprocessed, resampling_params, use_voxels=False, interp_type=0)
            if t1_segmentation is not None:
                t1_segmentation = ants.resample_image(t1_segmentation, resampling_params, use_voxels=False, interp_type=1)
            if brain_mask is not None:
                brain_mask = ants.resample_image(brain_mask, resampling_params, use_voxels=False, interp_type=1)

        ################################
        #
        # Build model and load weights
        #
        ################################

        template_size = (208, 208)

        image_modalities = ("T1", "FLAIR")
        if do_flair_only:
            image_modalities=("FLAIR",)
        elif do_t1_only:
            image_modalities=("T1",)
        if use_t1_segmentation:
            image_modalities = (*image_modalities, "T1Seg")
        channel_size = len(image_modalities)

        if which_model == "sysu":
            unet_model = create_unet_model_2d((*template_size, channel_size),
                number_of_outputs=1, mode="sigmoid",
                number_of_filters=(64, 96, 128, 256, 512), dropout_rate=0.0,
                convolution_kernel_size=(3, 3), deconvolution_kernel_size=(2, 2),
                weight_decay=0, additional_options=("initialConvolutionKernelSize[5]",))
        elif which_model == "sysuWithSite":
            unet_model = create_unet_model_2d((*template_size, channel_size),
                number_of_outputs=1, mode="sigmoid",
                scalar_output_size=3, scalar_output_activation="softmax",
                number_of_filters=(64, 96, 128, 256, 512), dropout_rate=0.0,
                convolution_kernel_size=(3, 3), deconvolution_kernel_size=(2, 2),
                weight_decay=0, additional_options=("initialConvolutionKernelSize[5]",))
        else:
            unet_model = create_unet_model_2d((*template_size, channel_size),
                number_of_outputs=1, mode="sigmoid",
                number_of_filters=(64, 96, 128, 256, 512), dropout_rate=0.0,
                convolution_kernel_size=(3, 3), deconvolution_kernel_size=(2, 2),
                weight_decay=1e-5,
                additional_options=("nnUnetActivationStyle", "attentionGating", "initialConvolutionKernelSize[5]",))

        if verbose == True:
            print("ewDavid:  retrieving model weights.")

        weights_file_name = None
        if which_model == "sysu" and flair is not None and t1 is not None:
            weights_file_name = get_pretrained_network("ewDavidSysu", antsxnet_cache_directory=antsxnet_cache_directory)
        elif which_model == "sysu" and flair is None and t1 is not None:
            weights_file_name = get_pretrained_network("ewDavidSysuT1Only", antsxnet_cache_directory=antsxnet_cache_directory)
        elif which_model == "sysu" and flair is not None and t1 is None:
            weights_file_name = get_pretrained_network("ewDavidSysuFlairOnly", antsxnet_cache_directory=antsxnet_cache_directory)
        elif which_model == "sysuPlus" and flair is not None and t1 is not None:
            weights_file_name = get_pretrained_network("ewDavidSysuPlus", antsxnet_cache_directory=antsxnet_cache_directory)
        elif which_model == "sysuPlus" and flair is None and t1 is not None:
            weights_file_name = get_pretrained_network("ewDavidSysuPlusT1Only", antsxnet_cache_directory=antsxnet_cache_directory)
        elif which_model == "sysuPlus" and flair is not None and t1 is None:
            weights_file_name = get_pretrained_network("ewDavidSysuPlusFlairOnly", antsxnet_cache_directory=antsxnet_cache_directory)
        elif which_model == "sysuPlusSeg" and flair is not None and t1 is not None:
            weights_file_name = get_pretrained_network("ewDavidSysuPlusSeg", antsxnet_cache_directory=antsxnet_cache_directory)
        elif which_model == "sysuPlusSeg" and flair is None and t1 is not None:
            weights_file_name = get_pretrained_network("ewDavidSysuPlusSegT1Only", antsxnet_cache_directory=antsxnet_cache_directory)
        elif which_model == "sysuWithSite" and flair is not None and t1 is not None:
            weights_file_name = get_pretrained_network("ewDavidSysuWithSite", antsxnet_cache_directory=antsxnet_cache_directory)
        elif which_model == "sysuWithSite" and flair is None and t1 is not None:
            weights_file_name = get_pretrained_network("ewDavidSysuWithSiteT1Only", antsxnet_cache_directory=antsxnet_cache_directory)
        elif which_model == "sysuWithSite" and flair is not None and t1 is None:
            weights_file_name = get_pretrained_network("ewDavidSysuWithSiteFlairOnly", antsxnet_cache_directory=antsxnet_cache_directory)
        else:
            raise ValueError("Incorrect model specification or image combination.")

        unet_model.load_weights(weights_file_name)

        ################################
        #
        # Data augmentation and extract slices
        #
        ################################

        wmh_probability_image = None
        if t1 is not None:
            wmh_probability_image = ants.image_clone(t1_preprocessed) * 0
        else:
            wmh_probability_image = ants.image_clone(flair_preprocessed) * 0

        wmh_site = np.array([0, 0, 0])

        data_augmentation = None
        if number_of_simulations > 0:
            if do_flair_only:
                data_augmentation = randomly_transform_image_data(
                    reference_image=flair_preprocessed,
                    input_image_list=[[flair_preprocessed]],
                    number_of_simulations=number_of_simulations,
                    transform_type='affine',
                    sd_affine=sd_affine,
                    input_image_interpolator='linear')
            elif do_t1_only:
                if use_t1_segmentation:
                    data_augmentation = randomly_transform_image_data(
                        reference_image=t1_preprocessed,
                        input_image_list=[[t1_preprocessed]],
                        segmentation_image_list=[t1_segmentation],
                        number_of_simulations=number_of_simulations,
                        transform_type='affine',
                        sd_affine=sd_affine,
                        input_image_interpolator='linear',
                        segmentation_image_interpolator='nearestNeighbor')
                else:
                    data_augmentation = randomly_transform_image_data(
                        reference_image=t1_preprocessed,
                        input_image_list=[[t1_preprocessed]],
                        number_of_simulations=number_of_simulations,
                        transform_type='affine',
                        sd_affine=sd_affine,
                        input_image_interpolator='linear')
            else:
                if use_t1_segmentation:
                    data_augmentation = randomly_transform_image_data(
                        reference_image=t1_preprocessed,
                        input_image_list=[[flair_preprocessed, t1_preprocessed]],
                        segmentation_image_list=[t1_segmentation],
                        number_of_simulations=number_of_simulations,
                        transform_type='affine',
                        sd_affine=sd_affine,
                        input_image_interpolator='linear',
                        segmentation_image_interpolator='nearestNeighbor')
                else:
                    data_augmentation = randomly_transform_image_data(
                        reference_image=t1_preprocessed,
                        input_image_list=[[flair_preprocessed, t1_preprocessed]],
                        number_of_simulations=number_of_simulations,
                        transform_type='affine',
                        sd_affine=sd_affine,
                        input_image_interpolator='linear')

        dimensions_to_predict = list((0,))
        if which_axes == "max":
            spacing = ants.get_spacing(wmh_probability_image)
            dimensions_to_predict = (spacing.index(max(spacing)),)
        elif which_axes == "all":
            dimensions_to_predict = list(range(3))
        else:
            if isinstance(which_axes, int):
                dimensions_to_predict = list((which_axes,))
            else:
                dimensions_to_predict = list(which_axes)

        total_number_of_slices = 0
        for d in range(len(dimensions_to_predict)):
            total_number_of_slices += wmh_probability_image.shape[dimensions_to_predict[d]]

        batchX = np.zeros((total_number_of_slices, *template_size, channel_size))

        for n in range(number_of_simulations + 1):

            batch_flair = flair_preprocessed
            batch_t1 = t1_preprocessed
            batch_t1_segmentation = t1_segmentation

            if n > 0:
                if do_flair_only:
                    batch_flair = data_augmentation['simulated_images'][n-1][0]
                elif do_t1_only:
                    batch_t1 = data_augmentation['simulated_images'][n-1][0]
                else:
                    batch_flair = data_augmentation['simulated_images'][n-1][0]
                    batch_t1 = data_augmentation['simulated_images'][n-1][1]
                if use_t1_segmentation:
                    batch_t1_segmentation = data_augmentation['simulated_segmentation_images'][n-1]

            if batch_flair is not None:
                batch_flair = (batch_flair - batch_flair[brain_mask != 0].mean()) / batch_flair[brain_mask != 0].std()
            if batch_t1 is not None:
                batch_t1 = (batch_t1 - batch_t1[brain_mask != 0].mean()) / batch_t1[brain_mask != 0].std()

            slice_count = 0
            for d in range(len(dimensions_to_predict)):

                number_of_slices = None
                if batch_t1 is not None:
                    number_of_slices = batch_t1.shape[dimensions_to_predict[d]]
                else:
                    number_of_slices = batch_flair.shape[dimensions_to_predict[d]]

                if verbose == True:
                    print("Extracting slices for dimension ", dimensions_to_predict[d])

                for i in range(number_of_slices):

                    channel_count = 0
                    if batch_flair is not None:
                        flair_slice = pad_or_crop_image_to_size(ants.slice_image(batch_flair, dimensions_to_predict[d], i), template_size)
                        batchX[slice_count,:,:,channel_count] = flair_slice.numpy()
                        channel_count += 1
                    if batch_t1 is not None:
                        t1_slice = pad_or_crop_image_to_size(ants.slice_image(batch_t1, dimensions_to_predict[d], i), template_size)
                        batchX[slice_count,:,:,channel_count] = t1_slice.numpy()
                        channel_count += 1
                    if t1_segmentation is not None:
                        t1_segmentation_slice = pad_or_crop_image_to_size(ants.slice_image(batch_t1_segmentation, dimensions_to_predict[d], i), template_size)
                        batchX[slice_count,:,:,channel_count] = t1_segmentation_slice.numpy() / 6

                    slice_count += 1

            ################################
            #
            # Do prediction and then restack into the image
            #
            ################################

            if verbose == True:
                if n == 0:
                    print("Prediction")
                else:
                    print("Prediction (simulation " + str(n) + ")")

            prediction = unet_model.predict(batchX, verbose=verbose)

            permutations = list()
            permutations.append((0, 1, 2))
            permutations.append((1, 0, 2))
            permutations.append((1, 2, 0))

            prediction_image_average = ants.image_clone(wmh_probability_image) * 0

            current_start_slice = 0
            for d in range(len(dimensions_to_predict)):
                current_end_slice = current_start_slice + wmh_probability_image.shape[dimensions_to_predict[d]]
                which_batch_slices = range(current_start_slice, current_end_slice)
                if isinstance(prediction, list):
                    prediction_per_dimension = prediction[0][which_batch_slices,:,:,0]
                else:
                    prediction_per_dimension = prediction[which_batch_slices,:,:,0]
                prediction_array = np.transpose(np.squeeze(prediction_per_dimension), permutations[dimensions_to_predict[d]])
                prediction_image = ants.copy_image_info(wmh_probability_image,
                    pad_or_crop_image_to_size(ants.from_numpy(prediction_array),
                    wmh_probability_image.shape))
                prediction_image_average = prediction_image_average + (prediction_image - prediction_image_average) / (d + 1)
                current_start_slice = current_end_slice

            wmh_probability_image = wmh_probability_image + (prediction_image_average - wmh_probability_image) / (n + 1)
            if isinstance(prediction, list):
                wmh_site = wmh_site + (np.mean(prediction[1], axis=0) - wmh_site) / (n + 1)

        if do_resampling:
            if t1 is not None:
                wmh_probability_image = ants.resample_image_to_target(wmh_probability_image, t1)
            if flair is not None:
                wmh_probability_image = ants.resample_image_to_target(wmh_probability_image, flair)

        if isinstance(prediction, list):
            return([wmh_probability_image, wmh_site])
        else:
            return(wmh_probability_image)




import tensorflow.keras.backend as K

from tensorflow.keras.models import Model
from tensorflow.keras.layers import Layer, InputSpec
from tensorflow.keras import initializers

import numpy as np

from ..architectures import create_autoencoder_model 
from ..architectures import create_convolutional_autoencoder_model_2d
from ..architectures import create_convolutional_autoencoder_model_3d

from sklearn.cluster import KMeans

class DeepEmbeddedClustering(Layer):

    """
    Deep embedded lustering layer.

    Arguments
    ---------
    number_of_clusters : integer
        Specifies which axis to normalize.

    initial_cluster_weights : list
        Initial clustering weights.

    alpha : scalar
        Parameter.

    Returns
    -------
    Keras layer
        A keras layer

    """

    def __init__(self, number_of_clusters=10, initial_cluster_weights=None, alpha=1.0, **kwargs):
        super(DeepEmbeddedClustering, self).__init__(**kwargs)

        self.number_of_clusters = number_of_clusters
        self.initial_cluster_weights = initial_cluster_weights
        self.alpha = alpha
        self.input_spec = InputSpec(ndim=2)

    def build(self, input_shape):
        self.input_spec = InputSpec(dtype=K.floatx(), shape=(None, input_shape[1]))

        if len(input_shape) != 2:
            raise ValueError("input_shape is not of length 2.")

        self.clusters = self.add_weight(shape=(self.number_of_clusters, input_shape[1]),
                                        initializer=initializers.glorot_uniform(),
                                        name='clusters')

        if self.initial_cluster_weights != None:
            self.set_weights(self.initial_cluster_weights)
            self.initial_cluster_weights = None

        self.built = True

    def call(self, inputs, mask=None):
        # Uses Student t-distribution (same as t-SNE)
        # inputs are the variable containing the data, shape=(number_of_samples, number_of_features)

        q = 1.0 / (1.0 + (K.sum(K.square(K.expand_dims(inputs, axis = 1)
            - self.clusters), axis=2) / self.alpha))
        q **= ((self.alpha + 1.0) / 2.0)
        q = K.transpose(K.transpose(q) / K.sum(q, axis=1))
        return(q)

    def compute_output_shape(self, input_shape):
        return(input_shape[0], self.number_of_clusters)

    def get_config(self):
        config = {"momentum": self.momentum, "axis": self.axis}
        base_config = super(DeepEmbeddedClustering, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

class DeepEmbeddedClusteringModel(object):
    """
    Deep embedded clustering with and without convolutions.

    Arguments
    ---------

    number_of_units_per_layer : integer
        Autoencoder number of units per layer.

    number_of_clusters : integer
        Number of clusters.

    alpha : scalar
        Parameter

    initializer : string
        Initializer for autoencoder.

    Returns
    -------
    Keras model
        A keras clustering model.

    """

    def __init__(self, number_of_units_per_layer=None, number_of_clusters=10, alpha=1.0,
                 initializer='glorot_uniform', convolutional=False, input_image_size=None):

        super(DeepEmbeddedClusteringModel, self).__init__()

        self.number_of_units_per_layer = number_of_units_per_layer
        self.number_of_clusters = number_of_clusters
        self.alpha = alpha
        self.initializer = initializer
        self.convolutional = convolutional
        self.input_image_size = input_image_size
        self.autoencoder = None
        self.encoder = None

        if self.convolutional == True:
            if self.input_image_size == None:
                raise ValueError("Need to specify the input image size for CNN.")

            if len(self.input_image_size) == 3:  # 2-D
                self.autoencoder, self.encoder = create_convolutional_autoencoder_model_2d(
                       input_image_size=self.input_image_size,
                       number_of_filters_per_layer=self.number_of_units_per_layer)
            else:
                self.autoencoder, self.encoder = create_convolutional_autoencoder_model_3d(
                       input_image_size=self.input_image_size,
                       number_of_filters_per_layer=self.number_of_units_per_layer)

        else:
            self.autoencoder, self.encoder = create_autoencoder_model(
                self.number_of_units_per_layer, initializer=self.initializer)

        clustering_layer = DeepEmbeddedClustering(
            self.number_of_clusters, name="clustering")(self.encoder.output) 

        if self.convolutional == True:
            self.model = Model(inputs=self.encoder.input, outputs=[clustering_layer, self.autoencoder.output])
        else:
            self.model = Model(inputs=self.encoder.input, outputs=clustering_layer)

    def pretrain(self, x, optimizer='adam', epochs=200, batch_size=256):
        self.autoencoder.compile(optimizer=optimizer, loss='mse')
        self.autoencoder.fit(x, x, batch_size=batch_size, epochs=epochs)

    def load_weights(self, weights):
        self.model.load_weights(weights)

    def extract_features(self, x):
        self.encoder.predict(x, verbose=0)

    def predict_cluster_labels(self, x):
        cluster_probabilities = self.model.predict(x, verbose=0)
        return(cluster_probabilities.argmax(1))

    def target_distribution(self, q):
        weight = q**2 / q.sum(0)
        p = (weight.T / weight.sum(1)).T
        return(p)

    def compile(self, optimizer='sgd', loss='kld', loss_weights=None):
        self.model.compile(optimizer=optimizer, loss=loss, loss_weights=loss_weights)

    def fit(self, x, max_number_of_iterations=2e4, batch_size=256, tolerance=1e-3, update_interval=140):

        # Initialize clusters using k-means

        kmeans = KMeans(n_clusters=self.number_of_clusters, n_init=20)
        current_prediction = kmeans.fit_predict(self.encoder.predict(x))
        previous_prediction = np.copy(current_prediction)
        self.model.get_layer(name='clustering').set_weights([kmeans.cluster_centers_])

        # Deep clustering

        loss = 100000000
        index = 0
        index_array = np.arange(x.shape[0])

        for i in range(max_number_of_iterations):

            if i % update_interval == 0:
                if self.convolutional == True:
                    q, _ = self.model.predict(x, verbose=0)
                else:
                    q = self.model.predict(x, verbose=0)                    
                p = self.target_distribution(q)

                # Met stopping criterion

                current_prediction = q.argmax(1)
                delta_label = (np.sum(current_prediction != previous_prediction).astype(np.float32) /
                               current_prediction.shape[0])
                previous_prediction = np.copy(current_prediction)

                print("Iteration ", i, " (out of ", max_number_of_iterations, 
                  "): loss = ", loss, ", delta_label = ", delta_label)

                if i > 0 and delta_label < tolerance:
                    break

            batch_indices = index_array[index * batch_size:min((index + 1) * batch_size, x.shape[0])]

            if self.convolutional == True:
                if len(self.input_image_size) == 3:
                    loss = self.model.train_on_batch(x=x[batch_indices,:,:,:], 
                                                     y=[p[batch_indices,:], x[batch_indices,:,:,:]])
                else:
                    loss = self.model.train_on_batch(x=x[batch_indices,:,:,:,:], 
                                                     y=[p[batch_indices,:,:], x[batch_indices,:,:,:,:]])
            else:
                loss = self.model.train_on_batch(x=x[batch_indices,:], y=p[batch_indices,:])

            if (index + 1) * batch_size <= x.shape[0]:
                index += 1
            else:
                index = 0

        return(current_prediction)




import tensorflow.keras.backend as K
from tensorflow.keras.layers import Layer, InputSpec
from tensorflow.keras import initializers

import tensorflow as tf

class SpatialTransformer2D(Layer):

    """
    Custom layer for the spatial transfomer network.

    Arguments
    ---------
    inputs : list of size 2
        The first element are the images and the second element are the
        weights.

    resampled_size : tuple of length 2
        Size of the resampled output images.

    transform_type : string
        Transform type (default = 'affine').

    interpolator_type : string
        Interpolator type (default = 'linear').

    Returns
    -------
    Keras layer
        A 2-D keras layer

    """

    def __init__(self, resampled_size, transform_type='affine', interpolator_type='linear', **kwargs):
        if K.backend() != 'tensorflow':
            raise ValueError("Tensorflow is required for this STN implementation.")
        if len(resampled_size) != 2:
            raise ValueError("Resampled size must be a vector of length 2 (for 2-D).")

        self.resampled_size = resampled_size
        self.transform_type = transform_type
        self.interpolator_type = interpolator_type

        super(SpatialTransformer2D, self).__init__(**kwargs)

    def build(self, input_shape):
        super(SpatialTransformer2D, self).build(input_shape)

    def call(self, inputs, mask=None):
        images = inputs[0]
        transform_parameters = inputs[1]

        output = None
        if self.transform_type == 'affine':
            output = self.affine_transform_images(images, transform_parameters, self.resampled_size)
        else:
            raise ValueError("Unsupported transform type.")

        input_shape = [K.shape(images)]
        return(tf.reshape(output, self.compute_output_shape(input_shape)))

    def compute_output_shape(self, input_shape):
        output_shape = (input_shape[0][0], self.resampled_size[0], self.resampled_size[1], input_shape[0][-1])
        return(output_shape)

    def affine_transform_images(self, images, affine_transform_parameters, resampled_size):
        batch_size = K.int_shape(images)[0]
        number_of_channels = K.int_shape(images)[-1]
        if batch_size is None:
            transform_parameters = K.reshape(affine_transform_parameters, shape=(-1, 2, 3))
        else:
            transform_parameters = K.reshape(affine_transform_parameters, shape=(batch_size, 2, 3))

        regular_grid = self.make_regular_grid(resampled_size)
        sampled_grids = K.dot(transform_parameters, regular_grid)

        if self.interpolator_type == 'linear':
            interpolated_image = self.linear_interpolate(images, sampled_grids, resampled_size)
        else:
            raise ValueError("Unsupported interpolator type.")

        if batch_size is None:
            new_output_shape = (-1, resampled_size[0], resampled_size[1], number_of_channels)
        else:
            new_output_shape = (batch_size, resampled_size[0], resampled_size[1], number_of_channels)

        interpolated_image = K.reshape(interpolated_image, shape = new_output_shape)

        return(interpolated_image)

    def make_regular_grid(self, resampled_size):
        x_linear_space = tf.linspace(-1.0, 1.0, resampled_size[1])
        y_linear_space = tf.linspace(-1.0, 1.0, resampled_size[0])

        x_coords, y_coords = tf.meshgrid(x_linear_space, y_linear_space)
        x_coords = K.flatten(x_coords)
        y_coords = K.flatten(y_coords)

        ones = K.ones_like(x_coords)
        regular_grid = K.concatenate([x_coords, y_coords, ones], axis = 0)
        regular_grid = K.flatten(regular_grid)
        regular_grid = K.reshape(regular_grid,
          (3, resampled_size[0] * resampled_size[1]))

        return(regular_grid)

    def linear_interpolate(self, images, sampled_grids, resampled_size):
        batch_size = K.shape(images)[0]
        height = K.shape(images)[1]
        width = K.shape(images)[2]
        number_of_channels = K.shape(images)[3]

        x = K.cast(K.flatten(sampled_grids[:, 0:1, :]), dtype='float32')
        y = K.cast(K.flatten(sampled_grids[:, 1:2, :]), dtype='float32')

        x = 0.5 * (x + 1.0) * K.cast(width, dtype='float32')
        y = 0.5 * (y + 1.0) * K.cast(height, dtype='float32')

        x0 = K.cast(x, dtype='int32')
        x1 = x0 + 1
        y0 = K.cast(y, dtype='int32')
        y1 = y0 + 1

        xMax = int(K.int_shape(images)[2] - 1)
        yMax = int(K.int_shape(images)[1] - 1)

        x0 = K.clip(x0, 0, xMax)
        x1 = K.clip(x1, 0, xMax)
        y0 = K.clip(y0, 0, yMax)
        y1 = K.clip(y1, 0, yMax)

        batch_pixels = K.arange(0, batch_size) * (height * width)
        batch_pixels = K.expand_dims(batch_pixels, axis = -1)
        base = K.repeat_elements(batch_pixels,
                                 rep=int(resampled_size[0] * resampled_size[1]), axis=1)
        base = K.flatten(base)

        indices00 = base + y0 * width + x0
        indices01 = base + y1 * width + x0
        indices10 = base + y0 * width + x1
        indices11 = base + y1 * width + x1

        flat_images = K.reshape(images, shape=(-1, number_of_channels))
        flat_images = K.cast(flat_images, dtype='float32')

        pixelValues00 = K.gather(flat_images, indices00)
        pixelValues01 = K.gather(flat_images, indices01)
        pixelValues10 = K.gather(flat_images, indices10)
        pixelValues11 = K.gather(flat_images, indices11)

        x0 = K.cast(x0, dtype='float32')
        x1 = K.cast(x1, dtype='float32')
        y0 = K.cast(y0, dtype='float32')
        y1 = K.cast(y1, dtype='float32')

        weight00 = K.expand_dims(((x1 - x) * (y1 - y)), axis=1)
        weight01 = K.expand_dims(((x1 - x) * (y - y0)), axis=1)
        weight10 = K.expand_dims(((x - x0) * (y1 - y)), axis=1)
        weight11 = K.expand_dims(((x - x0) * (y - y0)), axis=1)

        interpolatedValues00 = weight00 * pixelValues00
        interpolatedValues01 = weight01 * pixelValues01
        interpolatedValues10 = weight10 * pixelValues10
        interpolatedValues11 = weight11 * pixelValues11

        interpolatedValues = (interpolatedValues00 + interpolatedValues01 +
          interpolatedValues10 + interpolatedValues11)

        return(interpolatedValues)

class SpatialTransformer3D(Layer):

    """
    Custom layer for the spatial transfomer network.

    Arguments
    ---------
    inputs : list of size 2
        The first element are the images and the second element are the
        weights.

    resampled_size : tuple of length 3
        Size of the resampled output images.

    transform_type : string
        Transform type (default = 'affine').

    interpolator_type : string
        Interpolator type (default = 'linear').

    Returns
    -------
    Keras layer
        A 3-D keras layer

    """

    def __init__(self, resampled_size, transform_type='affine', interpolator_type='linear', **kwargs):
        if K.backend() != 'tensorflow':
            raise ValueError("Tensorflow is required for this STN implementation.")
        if len(resampled_size) != 3:
            raise ValueError("Resampled size must be a vector of length 3 (for 3-D).")

        self.resampled_size = resampled_size
        self.transform_type = transform_type
        self.interpolator_type = interpolator_type

        super(SpatialTransformer3D, self).__init__(**kwargs)

    def build(self, input_shape):
        super(SpatialTransformer3D, self).build(input_shape)

    def call(self, inputs, mask=None):
        images = inputs[0]
        transform_parameters = inputs[1]

        output = None
        if self.transform_type == 'affine':
            output = self.affine_transform_images(images, transform_parameters, self.resampled_size)
        else:
            raise ValueError("Unsupported transform type.")

        input_shape = [K.shape(images)]
        return(tf.reshape(output, self.compute_output_shape(input_shape)))

    def compute_output_shape(self, input_shape):
        output_shape = (input_shape[0][0], self.resampled_size[0],
                        self.resampled_size[1], self.resampled_size[2],
                        input_shape[0][-1])
        return(output_shape)

    def affine_transform_images(self, images, affine_transform_parameters, resampled_size):
        batch_size = K.int_shape(images)[0]
        number_of_channels = K.int_shape(images)[-1]
        if batch_size is None:
            transform_parameters = K.reshape(affine_transform_parameters, shape=(-1, 3, 4))
        else:
            transform_parameters = K.reshape(affine_transform_parameters, shape=(batch_size, 3, 4))

        regular_grid = self.make_regular_grid(resampled_size)
        sampled_grids = K.dot(transform_parameters, regular_grid)

        if self.interpolator_type == 'linear':
            interpolated_image = self.linear_interpolate(images, sampled_grids, resampled_size)
        else:
            raise ValueError("Unsupported interpolator type.")

        if batch_size is None:
            new_output_shape = (-1, resampled_size[0], resampled_size[1], resampled_size[2], number_of_channels)
        else:
            new_output_shape = (batch_size, resampled_size[0], resampled_size[1], resampled_size[2], number_of_channels)

        interpolated_image = K.reshape(interpolated_image, shape = new_output_shape)

        return(interpolated_image)

    def make_regular_grid(self, resampled_size):
        x_linear_space = tf.linspace(-1.0, 1.0, resampled_size[1])
        y_linear_space = tf.linspace(-1.0, 1.0, resampled_size[0])
        z_linear_space = tf.linspace(-1.0, 1.0, resampled_size[2])

        x_coords, y_coords, z_coords = tf.meshgrid(x_linear_space, y_linear_space, z_linear_space)
        x_coords = K.flatten(x_coords)
        y_coords = K.flatten(y_coords)
        z_coords = K.flatten(z_coords)

        ones = K.ones_like(x_coords)
        regular_grid = K.concatenate([x_coords, y_coords, z_coords, ones], axis = 0)
        regular_grid = K.flatten(regular_grid)
        regular_grid = K.reshape(regular_grid,
          (4, resampled_size[0] * resampled_size[1] * resampled_size[2]))

        return(regular_grid)

    def linear_interpolate(self, images, sampled_grids, resampled_size):
        batch_size = K.shape(images)[0]
        height = K.shape(images)[1]
        width = K.shape(images)[2]
        depth = K.shape(images)[3]
        number_of_channels = K.shape(images)[4]

        x = K.cast(K.flatten(sampled_grids[:, 0:1, :]), dtype='float32')
        y = K.cast(K.flatten(sampled_grids[:, 1:2, :]), dtype='float32')
        z = K.cast(K.flatten(sampled_grids[:, 2:3, :]), dtype='float32')

        x = 0.5 * (x + 1.0) * K.cast(width, dtype='float32')
        y = 0.5 * (y + 1.0) * K.cast(height, dtype='float32')
        z = 0.5 * (z + 1.0) * K.cast(depth, dtype='float32')

        x0 = K.cast(x, dtype='int32')
        x1 = x0 + 1
        y0 = K.cast(y, dtype='int32')
        y1 = y0 + 1
        z0 = K.cast(z, dtype='int32')
        z1 = z0 + 1

        xMax = int(K.int_shape(images)[2] - 1)
        yMax = int(K.int_shape(images)[1] - 1)
        zMax = int(K.int_shape(images)[3] - 1)

        x0 = K.clip(x0, 0, xMax)
        x1 = K.clip(x1, 0, xMax)
        y0 = K.clip(y0, 0, yMax)
        y1 = K.clip(y1, 0, yMax)
        z0 = K.clip(z0, 0, zMax)
        z1 = K.clip(z1, 0, zMax)

        batch_pixels = K.arange(0, batch_size) * (height * width * depth)
        batch_pixels = K.expand_dims(batch_pixels, axis=-1)
        base = K.repeat_elements(batch_pixels,
                                 rep=int(resampled_size[0] * resampled_size[1] * resampled_size[2]), axis=1)
        base = K.flatten(base)

        indices000 = base + z0 * (width * height) + y0 * width + x0
        indices001 = base + z1 * (width * height) + y0 * width + x0
        indices010 = base + z0 * (width * height) + y1 * width + x0
        indices011 = base + z1 * (width * height) + y1 * width + x0
        indices100 = base + z0 * (width * height) + y0 * width + x1
        indices101 = base + z1 * (width * height) + y0 * width + x1
        indices110 = base + z0 * (width * height) + y1 * width + x1
        indices111 = base + z1 * (width * height) + y1 * width + x1

        flatImages = K.reshape(images, shape=(-1, number_of_channels))
        flatImages = K.cast(flatImages, dtype='float32')

        pixelValues000 = K.gather(flatImages, indices000)
        pixelValues001 = K.gather(flatImages, indices001)
        pixelValues010 = K.gather(flatImages, indices010)
        pixelValues011 = K.gather(flatImages, indices011)
        pixelValues100 = K.gather(flatImages, indices100)
        pixelValues101 = K.gather(flatImages, indices101)
        pixelValues110 = K.gather(flatImages, indices110)
        pixelValues111 = K.gather(flatImages, indices111)

        x0 = K.cast(x0, dtype='float32')
        x1 = K.cast(x1, dtype='float32')
        y0 = K.cast(y0, dtype='float32')
        y1 = K.cast(y1, dtype='float32')
        z0 = K.cast(z0, dtype='float32')
        z1 = K.cast(z1, dtype='float32')

        weight000 = K.expand_dims(((x1 - x) * (y1 - y) * (z1 - z)), axis = 1)
        weight001 = K.expand_dims(((x1 - x) * (y1 - y) * (z - z0)), axis = 1)
        weight010 = K.expand_dims(((x1 - x) * (y - y0) * (z1 - z)), axis = 1)
        weight011 = K.expand_dims(((x1 - x) * (y - y0) * (z - z0)), axis = 1)
        weight100 = K.expand_dims(((x - x0) * (y1 - y) * (z1 - z)), axis = 1)
        weight101 = K.expand_dims(((x - x0) * (y1 - y) * (z - z0)), axis = 1)
        weight110 = K.expand_dims(((x - x0) * (y - y0) * (z1 - z)), axis = 1)
        weight111 = K.expand_dims(((x - x0) * (y - y0) * (z - z0)), axis = 1)

        interpolatedValues000 = weight000 * pixelValues000
        interpolatedValues001 = weight001 * pixelValues001
        interpolatedValues010 = weight010 * pixelValues010
        interpolatedValues011 = weight011 * pixelValues011
        interpolatedValues100 = weight100 * pixelValues100
        interpolatedValues101 = weight101 * pixelValues101
        interpolatedValues110 = weight110 * pixelValues110
        interpolatedValues111 = weight111 * pixelValues111

        interpolatedValues = (
          interpolatedValues000 +
          interpolatedValues001 +
          interpolatedValues010 +
          interpolatedValues011 +
          interpolatedValues100 +
          interpolatedValues101 +
          interpolatedValues110 +
          interpolatedValues111)

        return(interpolatedValues)














from tensorflow.keras.models import Model
from tensorflow.keras.layers import (Input, Average, Add,
                          Conv2D, Conv2DTranspose,
                          MaxPooling2D, UpSampling2D,
                          Conv3D, Conv3DTranspose,
                          MaxPooling3D, UpSampling3D)

import numpy as np

def create_deep_denoise_super_resolution_model_2d(input_image_size,
                                                  layers=2,
                                                  lowest_resolution=64,
                                                  convolution_kernel_size=(3, 3),
                                                  pool_size=(2, 2),
                                                  strides=(2, 2)
                                                 ):
    """
    2-D implementation of the denoising autoencoder image super resolution deep learning architecture.

    Arguments
    ---------
    input_image_size : tuple of length 3
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    layers : integer
        Number of architecture layers.

    lowest_resolution : integer
        Number of filters at the beginning and end of the architecture.

    convolution_kernel_size : 2-d tuple
        specifies the kernel size during the encoding path.

    pool_size : 2-d tuple
         Defines the region for each pooling layer.

    strides : 2-d tuple
         Defines the stride length in each direction.

    Returns
    -------
    Keras model
        A 2-D Keras model defining the network.

    Example
    -------
    >>> model = create_deep_denoise_super_resolution_model_2d((128, 128, 1))
    >>> model.summary()
    """

    inputs = Input(shape = input_image_size)

    # encoding layers

    pool = None
    conv = None

    encoding_convolution_layers = []
    for i in range(layers):
        number_of_filters = lowest_resolution * 2 ** i
        if i == 0:
            conv = Conv2D(filters=number_of_filters,
                          kernel_size=convolution_kernel_size,
                          activation='relu',
                          padding='same')(inputs)
        else:
            conv = Conv2D(filters=number_of_filters,
                          kernel_size=convolution_kernel_size,
                          activation='relu',
                          padding='same')(pool)

        layer = Conv2D(filters=number_of_filters,
                       kernel_size=convolution_kernel_size,
                       activation='relu',
                       padding='same')(conv)
        encoding_convolution_layers.append(layer)

        pool = MaxPooling2D(pool_size=pool_size,
                            strides=strides,
                            padding='same')(encoding_convolution_layers[i])

    number_of_filters = lowest_resolution * 2 ** layers

    outputs = Conv2D(filters=number_of_filters,
                     kernel_size=convolution_kernel_size,
                     activation='relu',
                     padding='same')(pool)

    # upsampling layers

    for i in range(layers):
        number_of_filters = lowest_resolution * 2 ** (layers - i - 1)

        outputs = UpSampling2D()(outputs)

        conv = Conv2D(filters=number_of_filters,
                      kernel_size=convolution_kernel_size,
                      activation='relu',
                      padding='same')(outputs)
        conv = Conv2D(filters=number_of_filters,
                      kernel_size=convolution_kernel_size,
                      activation='relu',
                      padding='same')(conv)

        outputs = Add()([encoding_convolution_layers[layers - i -1], conv])
        outputs = UpSampling2D()(outputs)

    number_of_channels = input_image_size[-1]

    final_convolution_kernel_size = tuple(np.array(convolution_kernel_size) + 2)

    outputs = Conv2D(filters=number_of_channels,
                     kernel_size=final_convolution_kernel_size,
                     activation='linear',
                     padding='same')(outputs)

    sr_model = Model(inputs=inputs, outputs=outputs)

    return(sr_model)


def create_deep_denoise_super_resolution_model_3d(input_image_size,
                                                  layers=2,
                                                  lowest_resolution=64,
                                                  convolution_kernel_size=(3, 3, 3),
                                                  pool_size=(2, 2, 2),
                                                  strides=(2, 2, 2)
                                                 ):
    """
    3-D implementation of the denoising autoencoder image super resolution deep learning architecture.

    Arguments
    ---------
    input_image_size : tuple of length 4
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    layers : integer
        Number of architecture layers.

    lowest_resolution : integer
        Number of filters at the beginning and end of the architecture.

    convolution_kernel_size : 3-d tuple
        specifies the kernel size during the encoding path.

    pool_size : 3-d tuple
         Defines the region for each pooling layer.

    strides : 3-d tuple
         Defines the stride length in each direction.

    Returns
    -------
    Keras model
        A 3-D Keras model defining the network.

    Example
    -------
    >>> model = create_deep_denoise_super_resolution_model_3d((128, 128, 128, 1))
    >>> model.summary()
    """

    inputs = Input(shape = input_image_size)

    # encoding layers

    pool = None
    conv = None

    encoding_convolution_layers = []
    for i in range(layers):
        number_of_filters = lowest_resolution * 2 ** i
        if i == 0:
            conv = Conv3D(filters=number_of_filters,
                          kernel_size=convolution_kernel_size,
                          activation='relu',
                          padding='same')(inputs)
        else:
            conv = Conv3D(filters=number_of_filters,
                          kernel_size=convolution_kernel_size,
                          activation='relu',
                          padding='same')(pool)

        layer = Conv3D(filters=number_of_filters,
                       kernel_size=convolution_kernel_size,
                       activation='relu',
                       padding='same')(conv)
        encoding_convolution_layers.append(layer)

        pool = MaxPooling3D(pool_size=pool_size,
                            strides=strides,
                            padding='same')(encoding_convolution_layers[i])

    number_of_filters = lowest_resolution * 2 ** layers

    outputs = Conv3D(filters=number_of_filters,
                     kernel_size=convolution_kernel_size,
                     activation='relu',
                     padding='same')(pool)

    # upsampling layers

    for i in range(layers):
        number_of_filters = lowest_resolution * 2 ** (layers - i - 1)

        outputs = UpSampling3D()(outputs)

        conv = Conv3D(filters=number_of_filters,
                      kernel_size=convolution_kernel_size,
                      activation='relu',
                      padding='same')(outputs)
        conv = Conv3D(filters=number_of_filters,
                      kernel_size=convolution_kernel_size,
                      activation='relu',
                      padding='same')(conv)

        outputs = Add()([encoding_convolution_layers[layers - i -1], conv])
        outputs = UpSampling3D()(outputs)

    number_of_channels = input_image_size[-1]

    final_convolution_kernel_size = tuple(np.array(convolution_kernel_size) + 2)

    outputs = Conv3D(filters=number_of_channels,
                     kernel_size=final_convolution_kernel_size,
                     activation='linear',
                     padding='same')(outputs)

    sr_model = Model(inputs=inputs, outputs=outputs)

    return(sr_model)





import tensorflow.keras.backend as K

from tensorflow.keras.layers import Layer, InputSpec

from tensorflow.keras import initializers

class Scale(Layer):

    """
    Custom layer used in the Dense U-net class for normalization which
    learns a set of weights and biases for scaling the input data.

    Arguments
    ---------

    axis : integer
        Specifies which axis to normalize.

    momentum : scalar
        Value used for computation of the exponential average of the
        mean and standard deviation.

    """

    def __init__(self, axis=-1, momentum=0.9, **kwargs):
        self.momentum = momentum
        self.axis = axis

        super(Scale, self).__init__(**kwargs)

    def build(self, input_shape):
        self.input_spec = [InputSpec(shape=input_shape)]
        output_shape = (int(input_shape[self.axis]),)

        gamma_initializer = initializers.Ones()
        beta_initializer = initializers.Zeros()

        self.gamma = K.variable(gamma_initializer(output_shape))
        self.beta = K.variable(beta_initializer(output_shape))
        self.trainable_weights = [self.gamma, self.beta]

    def call(self, inputs, mask=None):
        input_shape = self.input_spec[0].shape
        broadcast_shape = [1] * len(input_shape)
        broadcast_shape[self.axis] = input_shape[self.axis]

        output = (K.reshape(self.gamma, broadcast_shape) * inputs +
                  K.reshape(self.beta, broadcast_shape))
        return(output)

    def get_config(self):
        config = {"momentum": self.momentum, "axis": self.axis}
        base_config = super(Scale, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))







import tensorflow as tf

import tensorflow.keras.backend as K

from tensorflow.keras.models import Model
from tensorflow.keras.layers import (Input, Dropout, BatchNormalization, Add,
                                     LeakyReLU, Concatenate, Lambda, Dense,
                                     Reshape, Permute, Multiply,
                                     Conv2D, Conv2DTranspose,
                                     MaxPooling2D, GlobalAveragePooling2D,
                                     UpSampling2D,
                                     Conv3D, Conv3DTranspose,
                                     MaxPooling3D, GlobalAveragePooling3D,
                                     UpSampling3D)

def create_resnet_model_2d(input_image_size,
                           input_scalars_size=0,
                           number_of_classification_labels=1000,
                           layers=(1, 2, 3, 4),
                           residual_block_schedule=(3, 4, 6, 3),
                           lowest_resolution=64,
                           cardinality=1,
                           squeeze_and_excite=False,
                           mode='classification'
                          ):
    """
    2-D implementation of the ResNet deep learning architecture.

    Creates a keras model of the ResNet deep learning architecture for image
    classification.  The paper is available here:

            https://arxiv.org/abs/1512.03385

    This particular implementation was influenced by the following python
    implementation:

            https://gist.github.com/mjdietzx/0cb95922aac14d446a6530f87b3a04ce

    Arguments
    ---------
    input_image_size : tuple of length 3
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    input_scalars_size : integer
        Optional integer specifying the size of the input vector for scalars that
        get concatenated to the fully connected layer at the end of the network.

    number_of_classification_labels : integer
        Number of classification labels.

    layers : tuple
        A tuple determining the number of 'filters' defined at for each layer.

    residual_block_schedule : tuple
        A tuple defining the how many residual blocks repeats for each layer.

    lowest_resolution : integer
        Number of filters at the initial layer.

    cardinality : integer
        perform ResNet (cardinality = 1) or ResNeX (cardinality does not 1 but,
        instead, powers of 2---try '32').

    squeeze_and_excite : boolean
        add the squeeze-and-excite block variant.

    mode : string
        'classification' or 'regression'.  Default = 'classification'.

    Returns
    -------
    Keras model
        A 2-D Keras model defining the network.

    Example
    -------
    >>> model = create_resnet_model_2d((128, 128, 1))
    >>> model.summary()
    """

    def add_common_layers(model):
        model = BatchNormalization()(model)
        model = LeakyReLU()(model)
        return(model)

    def grouped_convolution_layer_2d(model,
                                     number_of_filters,
                                     strides):
        # Per standard ResNet, this is just a 2-D convolution
        if cardinality == 1:
            grouped_model = Conv2D(filters=number_of_filters,
                                   kernel_size=(3, 3),
                                   strides=strides,
                                   padding='same')(model)
            return(grouped_model)

        if number_of_filters % cardinality != 0:
            raise ValueError('number_of_filters `%` cardinality != 0')

        number_of_group_filters = int(number_of_filters / cardinality)

        convolution_layers = []
        for j in range(cardinality):
            local_layer = Lambda(lambda z: z[:, :, :,
              j * number_of_group_filters:j * number_of_group_filters + number_of_group_filters])(model)
            convolution_layers.append(Conv2D(filters=number_of_group_filters,
                                             kernel_size=(3, 3),
                                             strides=strides,
                                             padding='same')(local_layer))

        grouped_model = Concatenate()(convolution_layers)
        return(grouped_model)

    def squeeze_and_excite_block_2d(model,
                                    ratio=16):
        initial = model
        number_of_filters = K.int_shape(initial)[-1]

        block_shape = (1, 1, number_of_filters)

        block = GlobalAveragePooling2D()(initial)
        block = Reshape(target_shape=block_shape)(block)
        block = Dense(units=number_of_filters//ratio,
                      activation='relu',
                      kernel_initializer='he_normal',
                      use_bias=False)(block)
        block = Dense(units=number_of_filters,
                      activation='sigmoid',
                      kernel_initializer='he_normal',
                      use_bias=False)(block)

        x = Multiply()([initial, block])
        return(x)

    def residual_block_2d(model,
                         number_of_filters_in,
                         number_of_filters_out,
                         strides=(1, 1),
                         project_shortcut=False,
                         squeeze_and_excite=False):
        shortcut = model

        model = Conv2D(filters=number_of_filters_in,
                       kernel_size=(1, 1),
                       strides=(1, 1),
                       padding='same')(model)
        model = add_common_layers(model)

        # ResNeXt (identical to ResNet when `cardinality` == 1)
        model = grouped_convolution_layer_2d(model,
                                             number_of_filters=number_of_filters_in,
                                             strides=strides)
        model = add_common_layers(model)

        model = Conv2D(filters=number_of_filters_out,
                       kernel_size=(1, 1),
                       strides=(1, 1),
                       padding='same')(model)
        model = BatchNormalization()(model)

        if project_shortcut == True or strides != (1,1):
            shortcut = Conv2D(filters=number_of_filters_out,
                              kernel_size=(1, 1),
                              strides=strides,
                              padding='same')(shortcut)
            shortcut = BatchNormalization()(shortcut)

        if squeeze_and_excite == True:
            model = squeeze_and_excite_block_2d(model)

        model = Add()([shortcut, model])
        model = LeakyReLU()(model)
        return(model)


    input_image = Input(shape = input_image_size)

    n_filters = lowest_resolution

    outputs = Conv2D(filters=n_filters,
                     kernel_size=(7, 7),
                     strides=(2, 2),
                     padding='same')(input_image)
    outputs = add_common_layers(outputs)
    outputs = MaxPooling2D(pool_size=(3, 3),
                           strides=(2, 2),
                           padding='same')(outputs)

    for i in range(len(layers)):
        n_filters_in = lowest_resolution * 2**layers[i]
        n_filters_out = 2 * n_filters_in

        for j in range(residual_block_schedule[i]):
            project_shortcut = False
            if i == 0 and j == 0:
                project_shortcut = True

            if i > 0 and j == 0:
                strides = (2, 2)
            else:
                strides = (1, 1)

            outputs = residual_block_2d(outputs,
                                        number_of_filters_in=n_filters_in,
                                        number_of_filters_out=n_filters_out,
                                        strides=strides,
                                        project_shortcut=project_shortcut,
                                        squeeze_and_excite=squeeze_and_excite)

    outputs = GlobalAveragePooling2D()(outputs)

    layer_activation = ''
    if mode == 'classification':
        layer_activation = 'softmax'
    elif mode == 'regression':
        layer_activation = 'linear'
    else:
        raise ValueError('mode must be either `classification` or `regression`.')

    resnet_model = None
    if input_scalars_size > 0:
        input_scalars = Input( shape = (input_scalars_size,) )
        concatenated_layer = Concatenate()([outputs, input_scalars])
        outputs = Dense(units=number_of_classification_labels,
                        activation=layer_activation)(concatenated_layer)
        resnet_model = Model(inputs=[input_image, input_scalars], outputs = outputs)
    else:
        outputs = Dense(units=number_of_classification_labels,
                        activation=layer_activation)(outputs)
        resnet_model = Model(inputs=input_image, outputs=outputs)

    return(resnet_model)

def create_resnet_model_3d(input_image_size,
                           input_scalars_size=0,
                           number_of_classification_labels=1000,
                           layers=(1, 2, 3, 4),
                           residual_block_schedule=(3, 4, 6, 3),
                           lowest_resolution=64,
                           cardinality=1,
                           squeeze_and_excite=False,
                           mode='classification'
                          ):
    """
    3-D implementation of the ResNet deep learning architecture.

    Creates a keras model of the ResNet deep learning architecture for image
    classification.  The paper is available here:

            https://arxiv.org/abs/1512.03385

    This particular implementation was influenced by the following python
    implementation:

            https://gist.github.com/mjdietzx/0cb95922aac14d446a6530f87b3a04ce

    Arguments
    ---------
    input_image_size : tuple of length 4
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    input_scalars_size : integer
        Optional integer specifying the size of the input vector for scalars that
        get concatenated to the fully connected layer at the end of the network.

    number_of_classification_labels : integer
        Number of classification labels.

    layers : tuple
        A tuple determining the number of 'filters' defined at for each layer.

    residual_block_schedule : tuple
        A tuple defining the how many residual blocks repeats for each layer.

    lowest_resolution : integer
        Number of filters at the initial layer.

    cardinality : integer
        perform ResNet (cardinality = 1) or ResNeX (cardinality does not 1 but,
        instead, powers of 2---try '32').

    squeeze_and_excite : boolean
        add the squeeze-and-excite block variant.

    mode : string
        'classification' or 'regression'.  Default = 'classification'.

    Returns
    -------
    Keras model
        A 3-D Keras model defining the network.

    Example
    -------
    >>> model = create_resnet_model_3d((128, 128, 128, 1))
    >>> model.summary()
    """

    def add_common_layers(model):
        model = BatchNormalization()(model)
        model = LeakyReLU()(model)
        return(model)

    def grouped_convolution_layer_3d(model,
                                     number_of_filters,
                                     strides):
        # Per standard ResNet, this is just a 3-D convolution
        if cardinality == 1:
            grouped_model = Conv3D(filters=number_of_filters,
                                   kernel_size=(3, 3, 3),
                                   strides=strides,
                                   padding='same')(model)
            return(grouped_model)

        if number_of_filters % cardinality != 0:
            raise ValueError('number_of_filters `%` cardinality != 0')

        number_of_group_filters = int(number_of_filters / cardinality)

        convolution_layers = []
        for j in range(cardinality):
            local_layer = Lambda(lambda z: z[:, :, :, :,
              j * number_of_group_filters:j * number_of_group_filters + number_of_group_filters])(model)
            convolution_layers.append(Conv3D(filters=number_of_group_filters,
                                             kernel_size=(3, 3, 3),
                                             strides=strides,
                                             padding='same')(local_layer))

        grouped_model = Concatenate()(convolution_layers)
        return(grouped_model)

    def squeeze_and_excite_block_3d(model,
                                    ratio=16):
        initial = model
        number_of_filters = K.int_shape(initial)[-1]

        block_shape = (1, 1, 1, number_of_filters)

        block = GlobalAveragePooling3D()(initial)
        block = Reshape(target_shape=block_shape)(block)
        block = Dense(units=number_of_filters//ratio,
                      activation='relu',
                      kernel_initializer='he_normal',
                      use_bias=False)(block)
        block = Dense(units=number_of_filters,
                      activation='sigmoid',
                      kernel_initializer='he_normal',
                      use_bias=False)(block)

        x = Multiply()([initial, block])
        return(x)

    def residual_block_3d(model,
                          number_of_filters_in,
                          number_of_filters_out,
                          strides=(1, 1, 1),
                          project_shortcut=False,
                          squeeze_and_excite=False):
        shortcut = model

        model = Conv3D(filters=number_of_filters_in,
                       kernel_size=(1, 1, 1),
                       strides=(1, 1, 1),
                       padding='same')(model)
        model = add_common_layers(model)

        # ResNeXt (identical to ResNet when `cardinality` == 1)
        model = grouped_convolution_layer_3d(model,
                                             number_of_filters=number_of_filters_in,
                                             strides=strides)
        model = add_common_layers(model)

        model = Conv3D(filters=number_of_filters_out,
                       kernel_size=(1, 1, 1),
                       strides=(1, 1, 1),
                       padding='same')(model)
        model = BatchNormalization()(model)

        if project_shortcut == True or strides != (1,1,1):
            shortcut = Conv3D(filters=number_of_filters_out,
                              kernel_size=(1, 1, 1),
                              strides=strides,
                              padding='same')(shortcut)
            shortcut = BatchNormalization()(shortcut)

        if squeeze_and_excite == True:
            model = squeeze_and_excite_block_3d(model)

        model = Add()([shortcut, model])
        model = LeakyReLU()(model)
        return(model)


    input_image = Input(shape = input_image_size)

    n_filters = lowest_resolution

    outputs = Conv3D(filters=n_filters,
                     kernel_size=(7, 7, 7),
                     strides=(2, 2, 2),
                     padding='same')(input_image)
    outputs = add_common_layers(outputs)
    outputs = MaxPooling3D(pool_size=(3, 3, 3),
                           strides=(2, 2, 2),
                           padding='same')(outputs)

    for i in range(len(layers)):
        n_filters_in = lowest_resolution * 2**layers[i]
        n_filters_out = 2 * n_filters_in

        for j in range(residual_block_schedule[i]):
            project_shortcut = False
            if i == 0 and j == 0:
                project_shortcut = True

            if i > 0 and j == 0:
                strides = (2, 2, 2)
            else:
                strides = (1, 1, 1)

            outputs = residual_block_3d(outputs,
                                        number_of_filters_in=n_filters_in,
                                        number_of_filters_out=n_filters_out,
                                        strides=strides,
                                        project_shortcut=project_shortcut,
                                        squeeze_and_excite=squeeze_and_excite)


    outputs = GlobalAveragePooling3D()(outputs)

    layer_activation = ''
    if mode == 'classification':
        layer_activation = 'softmax'
    elif mode == 'regression':
        layer_activation = 'linear'
    else:
        raise ValueError('mode must be either `classification` or `regression`.')

    resnet_model = None
    if input_scalars_size > 0:
        input_scalars = Input( shape = (input_scalars_size,) )
        concatenated_layer = Concatenate()([outputs, input_scalars])
        outputs = Dense(units=number_of_classification_labels,
                        activation=layer_activation)(concatenated_layer)
        resnet_model = Model(inputs=[input_image, input_scalars], outputs = outputs)
    else:
        outputs = Dense(units=number_of_classification_labels,
                        activation=layer_activation)(outputs)
        resnet_model = Model(inputs=input_image, outputs=outputs)

    return(resnet_model)







import tensorflow as tf

import tensorflow.keras.backend as K

from tensorflow.keras.models import Model
from tensorflow.keras.layers import (Input, Dropout, BatchNormalization, Add,
                                     Activation, Dense, Flatten,
                                     Conv2D, Conv2DTranspose, AveragePooling2D,
                                     Conv3D, Conv3DTranspose, AveragePooling3D)

def create_wide_resnet_model_2d(input_image_size,
                                number_of_classification_labels=1000,
                                depth=2,
                                width=1,
                                residual_block_schedule=(16, 32, 64),
                                pool_size=(8, 8),
                                dropout_rate=0.0,
                                weight_decay=0.0005,
                                mode='classification'
                               ):
    """
    2-D implementation of the Wide ResNet deep learning architecture.

    Creates a keras model of the Wide ResNet deep learning architecture for image
    classification/regression.  The paper is available here:

            https://arxiv.org/abs/1512.03385

    This particular implementation was influenced by the following python
    implementation:

            https://github.com/titu1994/Wide-Residual-Networks

    Arguments
    ---------
    input_image_size : tuple of length 3
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    number_of_classification_labels : integer
        Number of classification labels.

    depth : integer
        Determines the depth of the network.  Related to the actual number of
        layers by number_of_layers = depth * 6 + 4.    Default = 2 (such that
        number_of_layers = 16).

    width : integer
        Determines the width of the network.  Default = 1.

    residual_block_schedule : tuple
        Determines the number of filters per convolutional block. Default =
        (16, 32, 64).

    pool_size : tuple
        Pool size for final average pooling layer.  Default = (8, 8).

    dropout_rate : scalar
        Float between 0 and 1 to use between dense layers.

    weight_decay :  scalar
        Weighting parameter for regularization of the kernel weights of the
        convolution layers.  Default = 0.0005.

    mode : string
        'classification' or 'regression'.  Default = 'classification'.

    Returns
    -------
    Keras model
        A 2-D Keras model defining the network.

    Example
    -------
    >>> model = create_wide_resnet_model_2d((128, 128, 1))
    >>> model.summary()
    """

    channel_axis = 0
    if K.image_data_format() == 'channels_last':
        channel_axis = -1

    def initial_convolution_layer(model, number_of_filters):
        model = Conv2D(filters=number_of_filters,
                       kernel_size=(3, 3),
                       padding='same',
                       kernel_initializer=initializers.he_normal(),
                       kernel_regularizer=regularizers.l2(weight_decay))(model)
        model = BatchNormalization(axis=channel_axis,
                                   momentum=0.1,
                                   epsilon=1e-5,
                                   gamma_initializer="uniform")(model)
        model = Activation(activation='relu')(model)

        return(model)

    def custom_convolution_layer(initial_model, base, width, strides=(1, 1),
                                 dropout_rate=0.0, expand=True):
        number_of_filters=int(base * width)

        if expand:
            model = Conv2D(filters=number_of_filters,
                           kernel_size=(3, 3),
                           padding='same',
                           strides=strides,
                           use_bias=False,
                           kernel_initializer=initializers.he_normal(),
                           kernel_regularizer=regularizers.l2(weight_decay))(initial_model)
        else:
            model = initial_model

        model = BatchNormalization(axis=channel_axis,
                                   momentum=0.1,
                                   epsilon=1e-5,
                                   gamma_initializer="uniform")(model)
        model = Activation(activation='relu')(model)

        model = Conv2D(filters=number_of_filters,
                       kernel_size=(3, 3),
                       padding='same',
                       use_bias=False,
                       kernel_initializer=initializers.he_normal(),
                       kernel_regularizer=regularizers.l2(weight_decay))(model)

        if expand:
            skip_layer = Conv2D(filters=number_of_filters,
                                kernel_size=(1, 1),
                                padding='same',
                                strides=strides,
                                use_bias=False,
                                kernel_initializer=initializers.he_normal(),
                                kernel_regularizer=regularizers.l2(weight_decay))(initial_model)
            model = Add()([model, skip_layer])
        else:
            if dropout_rate > 0.0:
                model = Dropout(rate=dropout_rate)(model)

            model = BatchNormalization(axis=channel_axis,
                                       momentum=0.1,
                                       epsilon=1e-5,
                                       gamma_initializer="uniform")(model)
            model = Activation(activation='relu')(model)

            model = Conv2D(filters=number_of_filters,
                           kernel_size=(3, 3),
                           padding='same',
                           use_bias=False,
                           kernel_initializer=initializers.he_normal(),
                           kernel_regularizer=regularizers.l2(weight_decay))(model)

            model = Add()([initial_model, model])

        return(model)

    inputs = Input(shape=input_image_size)

    outputs = initial_convolution_layer(inputs, residual_block_schedule[0])
    number_of_convolutions = 4

    for i in range(len(residual_block_schedule)):
        base_number_of_filters = residual_block_schedule[i]

        outputs = custom_convolution_layer(outputs, base = base_number_of_filters,
                                           width=width, strides=(1, 1), dropout_rate=0.0,
                                           expand=True)
        number_of_convolutions += 2

        for j in range(depth):
            outputs = custom_convolution_layer(outputs, base = base_number_of_filters,
                                               width=width, dropout_rate=dropout_rate,
                                               expand=False)
            number_of_convolutions += 2

        outputs = BatchNormalization(axis=channel_axis,
                                     momentum=0.1,
                                     epsilon=1e-5,
                                     gamma_initializer="uniform")(outputs)
        outputs = Activation(activation='relu')(outputs)

    outputs = AveragePooling2D(pool_size=pool_size)(outputs)
    outputs = Flatten()(outputs)

    layer_activation = ''
    if mode == 'classification':
        layer_activation = 'softmax'
    elif mode == 'regression':
        layerActivation = 'linear'
    else:
        raise ValueError('mode must be either `classification` or `regression`.')

    outputs = Dense(units=number_of_classification_labels,
                    activation=layer_activation,
                    kernel_regularizer=regularizers.l2(weight_decay))(outputs)

    wide_resnet_model = Model(inputs=inputs, outputs=outputs)

    return(wide_resnet_model)


def create_wide_resnet_model_3d(input_image_size,
                                number_of_classification_labels=1000,
                                depth=2,
                                width=1,
                                residual_block_schedule=(16, 32, 64),
                                pool_size=(8, 8, 8),
                                dropout_rate=0.0,
                                weight_decay=0.0005,
                                mode='classification'
                               ):
    """
    3-D implementation of the Wide ResNet deep learning architecture.

    Creates a keras model of the Wide ResNet deep learning architecture for image
    classification/regression.  The paper is available here:

            https://arxiv.org/abs/1512.03385

    This particular implementation was influenced by the following python
    implementation:

            https://github.com/titu1994/Wide-Residual-Networks

    Arguments
    ---------
    input_image_size : tuple of length 4
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    number_of_classification_labels : integer
        Number of classification labels.

    depth : integer
        Determines the depth of the network.  Related to the actual number of
        layers by number_of_layers = depth * 6 + 4.    Default = 2 (such that
        number_of_layers = 16).

    width : integer
        Determines the width of the network.  Default = 1.

    residual_block_schedule : tuple
        Determines the number of filters per convolutional block. Default =
        (16, 32, 64).

    pool_size : tuple
        Pool size for final average pooling layer.  Default = (8, 8, 8).

    dropout_rate : scalar
        Float between 0 and 1 to use between dense layers.

    weight_decay :  scalar
        Weighting parameter for regularization of the kernel weights of the
        convolution layers.  Default = 0.0005.

    mode : string
        'classification' or 'regression'.  Default = 'classification'.

    Returns
    -------
    Keras model
        A 3-D Keras model defining the network.

    Example
    -------
    >>> model = create_wide_resnet_model_3d((128, 128, 128, 1))
    >>> model.summary()
    """

    channel_axis = 0
    if K.image_data_format() == 'channels_last':
        channel_axis = -1

    def initial_convolution_layer(model, number_of_filters):
        model = Conv3D(filters=number_of_filters,
                       kernel_size=(3, 3, 3),
                       padding='same',
                       kernel_initializer=initializers.he_normal(),
                       kernel_regularizer=regularizers.l2(weight_decay))(model)
        model = BatchNormalization(axis=channel_axis,
                                   momentum=0.1,
                                   epsilon=1e-5,
                                   gamma_initializer="uniform")(model)
        model = Activation(activation='relu')(model)

        return(model)

    def custom_convolution_layer(initial_model, base, width, strides=(1, 1, 1),
                                 dropout_rate=0.0, expand=True):
        number_of_filters=int(base * width)

        if expand:
            model = Conv3D(filters=number_of_filters,
                           kernel_size=(3, 3, 3),
                           padding='same',
                           strides=strides,
                           use_bias=False,
                           kernel_initializer=initializers.he_normal(),
                           kernel_regularizer=regularizers.l2(weight_decay))(initial_model)
        else:
            model = initial_model

        model = BatchNormalization(axis=channel_axis,
                                   momentum=0.1,
                                   epsilon=1e-5,
                                   gamma_initializer="uniform")(model)
        model = Activation(activation='relu')(model)

        model = Conv3D(filters=number_of_filters,
                       kernel_size=(3, 3, 3),
                       padding='same',
                       use_bias=False,
                       kernel_initializer=initializers.he_normal(),
                       kernel_regularizer=regularizers.l2(weight_decay))(model)

        if expand:
            skip_layer = Conv3D(filters=number_of_filters,
                                kernel_size=(1, 1, 1),
                                padding='same',
                                strides=strides,
                                use_bias=False,
                                kernel_initializer=initializers.he_normal(),
                                kernel_regularizer=regularizers.l2(weight_decay))(initial_model)
            model = Add()([model, skip_layer])
        else:
            if dropout_rate > 0.0:
                model = Dropout(rate=dropout_rate)(model)

            model = BatchNormalization(axis=channel_axis,
                                       momentum=0.1,
                                       epsilon=1e-5,
                                       gamma_initializer="uniform")(model)
            model = Activation(activation='relu')(model)

            model = Conv3D(filters=number_of_filters,
                           kernel_size=(3, 3, 3),
                           padding='same',
                           use_bias=False,
                           kernel_initializer=initializers.he_normal(),
                           kernel_regularizer=regularizers.l2(weight_decay))(model)

            model = Add()([initial_model, model])

        return(model)

    inputs = Input(shape=input_image_size)

    outputs = initial_convolution_layer(inputs, residual_block_schedule[0])
    number_of_convolutions = 4

    for i in range(len(residual_block_schedule)):
        base_number_of_filters = residual_block_schedule[i]

        outputs = custom_convolution_layer(outputs, base = base_number_of_filters,
                                           width=width, strides=(1, 1, 1), dropout_rate=0.0,
                                           expand=True)
        number_of_convolutions += 2

        for j in range(depth):
            outputs = custom_convolution_layer(outputs, base = base_number_of_filters,
                                               width=width, dropout_rate=dropout_rate,
                                               expand=False)
            number_of_convolutions += 2

        outputs = BatchNormalization(axis=channel_axis,
                                     momentum=0.1,
                                     epsilon=1e-5,
                                     gamma_initializer="uniform")(outputs)
        outputs = Activation(activation='relu')(outputs)

    outputs = AveragePooling3D(pool_size=pool_size)(outputs)
    outputs = Flatten()(outputs)

    layer_activation = ''
    if mode == 'classification':
        layer_activation = 'softmax'
    elif mode == 'regression':
        layerActivation = 'linear'
    else:
        raise ValueError('mode must be either `classification` or `regression`.')

    outputs = Dense(units=number_of_classification_labels,
                    activation=layer_activation,
                    kernel_regularizer=regularizers.l2(weight_decay))(outputs)

    wide_resnet_model = Model(inputs=inputs, outputs=outputs)

    return(wide_resnet_model)

